####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Application.

from fiction_dl.Concepts.Chapter import Chapter
from fiction_dl.Concepts.Extractor import Extractor
from fiction_dl.Utilities.HTML import StripHTML
import fiction_dl.Configuration as Configuration

# Standard packages.

from datetime import datetime
import logging
import re
from typing import List, Optional

# Non-standard packages.

from bs4 import BeautifulSoup
from dreamy_utilities.Text import GetCurrentDate, Stringify
from dreamy_utilities.Web import GetHostname, GetSiteURL
from dreamy_utilities.WebSession import WebSession

#
#
#
# The class definition.
#
#
#

class ExtractorFFNet(Extractor):


    def __init__(self) -> None:

        ##
        #
        # The constructor.
        #
        ##

        super().__init__()

        self._webSession.EnableCloudscraper(True)
        self._chapterParserName = "html5lib"

    def GetSupportedHostnames(self) -> List[str]:

        ##
        #
        # Returns a list of hostnames supposed to be supported by the extractor.
        #
        # @return A list of supported hostnames.
        #
        ##

        return [
            "fanfiction.net",
            "fictionpress.com"
        ]

    def ScanChannel(self, URL: str) -> Optional[List[str]]:

        ##
        #
        # Scans the channel: generates the list of story URLs.
        #
        # @return **None** when the scan fails, a list of story URLs when it doesn't fail.
        #
        ##

        if (not URL) or (GetHostname(URL) not in self.GetSupportedHostnames()):
            return None

        elif "/community/" in URL:
            return self._ScanCollection(URL)

        userIDMatch = re.search("/u/(\d+)", URL)
        if not userIDMatch:
            return None

        userID = userIDMatch.group(1)

        siteURL = GetSiteURL(URL)
        normalizedURL = f"{siteURL}/u/{userID}/"

        pageSoup = self._webSession.GetSoup(normalizedURL)
        if not pageSoup:
            return None

        storyIDs = []

        storyElements = pageSoup.find_all("div", {"class": "mystories"})
        for element in storyElements:

            linkElement = element.find("a", {"class": "stitle"})
            if (not linkElement) or (not linkElement.has_attr("href")):
                logging.error("Failed to retrieve story URL.")
                continue

            storyIDMatch = re.search("/s/(\d+)", linkElement["href"])
            if not storyIDMatch:
                logging.error("Failed to retrieve story ID from its URL.")
                continue

            storyID = storyIDMatch.group(1)
            storyIDs.append(storyID)

        storyURLs = [f"{siteURL}/s/{ID}/" for ID in storyIDs]
        return storyURLs

    def _ScanCollection(self, URL: str) -> Optional[List[str]]:

        ##
        #
        # Scans the channel: generates the list of story URLs.
        #
        # @return **None** when the scan fails, a list of story URLs when it doesn't fail.
        #
        ##

        # Retrieve collection name and generate a normalized URL.

        collectionNameAndIDMatch = re.search(
            "/community/([a-zA-Z0-9-]+)/(\d+)",
            URL
        )
        if not collectionNameAndIDMatch:
            logging.error("Failed to retrieve collection name/ID.")
            return None

        collectionName = collectionNameAndIDMatch.group(1)
        collectionID = collectionNameAndIDMatch.group(2)

        siteURL = GetSiteURL(URL)
        collectionURL = f"{siteURL}/community/{collectionName}/{collectionID}"
        normalizedURL = f"{collectionURL}/99/0/1/0/0/0/0/"

        # Download the first page.

        soup = self._webSession.GetSoup(normalizedURL)
        if not soup:
            logging.error(f"Failed to download page: \"{normalizedURL}\".")
            return None

        # Retrieve the number of pages.

        lastPageIndex = 1
        lastPageRelativeURL = None

        for elementCandidate in soup.select("center > a"):

            text = elementCandidate.get_text().strip()

            if "Last" == text:
                lastPageRelativeURL = elementCandidate["href"]
                break

        if lastPageRelativeURL:

            lastPageURLParts = lastPageRelativeURL.split("/")

            if len(lastPageURLParts) > 8:
                lastPageIndex = int(lastPageURLParts[-6])

        # Process each page of the collection.

        storyIDs = []

        for pageIndex in range(1, lastPageIndex + 1):

            pageURL = f"{collectionURL}/99/0/{pageIndex}/0/0/0/0/"
            soup = self._webSession.GetSoup(pageURL)
            if not soup:
                logging.error(f"Failed to download page: \"{pageURL}\".")
                return None

            for element in soup.select("div.z-list"):

                anchorElement = element.select_one("a.stitle")
                if (not anchorElement) or (not anchorElement.has_attr("href")):
                    logging.error("Failed to retrieve story URL.")
                    continue

                storyIDMatch = re.search("/s/(\d+)", anchorElement["href"])
                if not storyIDMatch:
                    logging.error("Failed to retrieve story ID from its URL.")
                    continue

                storyID = storyIDMatch.group(1)
                storyIDs.append(storyID)

        # Return.

        storyURLs = [f"{siteURL}/s/{ID}/" for ID in storyIDs]
        return storyURLs

    def _InternallyScanStory(
        self,
        URL: str,
        soup: Optional[BeautifulSoup]
    ) -> bool:

        ##
        #
        # Scans the story: generates the list of chapter URLs and retrieves the
        # metadata.
        #
        # @param URL  The URL of the story.
        # @param soup The tag soup.
        #
        # @return **False** when the scan fails, **True** when it doesn't fail.
        #
        ##

        # Extract metadata.

        headerElement = soup.find(id = "profile_top")
        if not headerElement:
            logging.error("Header element not found.")
            return False

        headerLines = headerElement.get_text().replace("Follow/Fav", "").split("\n")

        chapterCount = re.search("Chapters: (\d+)", headerLines[3])
        # If the story has just one chapter, this field won't be present.

        words = re.search("Words: ([\d,]+)", headerLines[3])
        if not words:
            logging.error("Word count field not found in header.")
            return False

        datePublished = re.search("Published: ([\d//]+)", headerLines[3])
        if not datePublished:
            logging.error("Date published field not found in header.")
            return False

        dateUpdated = re.search("Updated: ([\d//]+)", headerLines[3])
        # If the story has just one chapter, this field won't be present.

        # Set the metadata.

        self.Story.Metadata.Title = headerLines[0].strip()
        self.Story.Metadata.Author = headerLines[1][4:].strip() # Removes the "By: " part.
        self.Story.Metadata.Summary = StripHTML(headerLines[2]).strip()

        self.Story.Metadata.DatePublished = self._ReformatDate(datePublished.group(1))
        self.Story.Metadata.DateUpdated = (
            self._ReformatDate(dateUpdated.group(1))
            if dateUpdated else
            self.Story.Metadata.DatePublished
        )

        self.Story.Metadata.ChapterCount = int(chapterCount.group(1)) if chapterCount else 1
        self.Story.Metadata.WordCount = int(words.group(1).replace(",", ""))

        # Retrieve chapter URLs.

        storyID = self._GetStoryID(self.Story.Metadata.URL)
        if not storyID:
            logging.error("Failed to retrieve story ID from URL.")
            return False

        baseURL = GetSiteURL(self.Story.Metadata.URL)
        for index in range(1, self.Story.Metadata.ChapterCount + 1):
            self._chapterURLs.append(f"{baseURL}/s/{storyID}/{index}/")

        # Return.

        return True

    def _InternallyExtractChapter(
        self,
        URL: str,
        soup: Optional[BeautifulSoup]
    ) -> Optional[Chapter]:

        ##
        #
        # Extracts specific chapter.
        #
        # @param URL  The URL of the page containing the chapter.
        # @param soup The tag soup of the page containing the chapter.
        #
        # @return **True** if the chapter is extracted correctly, **False** otherwise.
        #
        ##

        # Read the title.

        title = None

        if (selectedChapterElement := soup.find("option", {"selected": True})):
            title = selectedChapterElement.text.strip()

        if title and (titleMatch := re.search("\d+\. (.*)", title)):
            title = titleMatch.group(1)

        # Read the content.

        storyTextElement = soup.find(id = "storytext")
        if not storyTextElement:
            logging.error("Story text element not found.")
            return None

        # Create the Chapter and return it.

        return Chapter(
            title = title,
            content = Stringify(storyTextElement.encode_contents())
        )

    @staticmethod
    def _GetStoryID(URL: str) -> Optional[str]:

        if not URL:
            return None

        storyIDMatch = re.search("/s/(\d+)/", URL)
        if not storyIDMatch:
            return None

        return storyIDMatch.group(1)

    @staticmethod
    def _ReformatDate(date: str) -> Optional[str]:

        if not date:
            return None

        try:

            # Dates on FF.net come in two formats: "m/d/yyyy" and "m/d".
            # We want to convert the shorter format to the longer one.

            longDatePattern = re.compile("^\d+/\d+/\d+$")

            if not longDatePattern.match(date):
                date += f"/{datetime.now().year}"

            # And then the longer format to the standard format.

            date = datetime.strptime(date, "%m/%d/%Y").strftime("%Y-%m-%d")

            return date

        except ValueError:

            return GetCurrentDate()


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Application.

from fiction_dl.Concepts.Processor import Processor
from fiction_dl.Utilities.HTML import CleanHTML, StripEmptyTags, StripTags

# Standard packages.

import re
from typing import Optional

# Non-standard packages.

from bs4 import  BeautifulSoup

#
#
#
# Classes.
#
#
#

##
#
# Santizes story content, i.e. prepares it to be used by a Formatter.
#
##

class SanitizerProcessor(Processor):

    def Process(self, content: str) -> Optional[str]:

        ##
        #
        # Processes the content given and returns a sanitized version of it.
        #
        # Sanitized code is defined as HTML code that contains no other tags than "p", "img", "hr",
        # "b", "strong", "i", "em", "u", and "a". No tag is empty, other than "img" and "hr". No tag
        # containes any attributes, with the exception of "a" - and the only attribute this tag is
        # allowed to have is "href".
        #
        # @param content The content to be processed.
        #
        # @return The processed content.
        #
        ##

        if not content:
            return None

        # Fix line breaks.

        content = self._FixLineBreaks(content)

        # Clean the code a bit and make sure it contains only allowed tags.

        content = CleanHTML(content)
        content = StripTags(content, self._Tags)

        # Strip empty tags.

        content = StripEmptyTags(content, ["hr", "img"])

        # Strip attributes.

        soup = BeautifulSoup(content, features = "html.parser")

        for tag in soup.find_all(lambda x: len(x.attrs) > 0):

            for name, value in list(tag.attrs.items()):

                if ("a" == tag.name) and ("href" == name):

                    # Fix parentheses - Typography Processor messes the up.

                    if value.startswith("“"):
                        value = value[1:]

                    if value.endswith("”"):
                        value = value[:-1]

                    tag[name] = value

                    continue

                else:

                    del tag[name]

        content = str(soup)

        # Strip newlines.

        Newlines = [
            "\r\n",
            "\r",
            "\n"
        ]

        for newline in Newlines:
            content = content.replace(newline, " ")

        # Return processed content.

        return content

    @staticmethod
    def _FixLineBreaks(content: str) -> Optional[str]:

        ##
        #
        # Replaces primitive line breaks ("<br>", "<br/>") with paragraphs.
        #
        # @param content Input content.
        #
        # @return Processed content, optionally **None**.
        #
        ##

        if not content:
            return None

        if ("<br>" in content) or ("<br/>" in content):

            if "<p>" not in content:
                content = "<p>" + content + "</p>"

            content = content.replace("<br>", "</p><p>")
            content = content.replace("<br/>", "</p><p>")

        return content

    _Tags = ["p", "img", "hr", "b", "strong", "i", "em", "u", "a"]


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Application.

from fiction_dl.Concepts.Chapter import Chapter
from fiction_dl.Concepts.Extractor import Extractor
from fiction_dl.Utilities.Filesystem import GetPackageDirectory
from fiction_dl.Utilities.Text import GetTitleProper
import fiction_dl.Configuration as Configuration

# Standard packages.

import logging
from random import randint
import re
import socket
from typing import List, Optional

# Non-standard packages.

from bs4 import BeautifulSoup
from dreamy_utilities.Filesystem import ReadTextFile
from dreamy_utilities.Interface import Interface
from dreamy_utilities.Text import GetDateFromTimestamp, GetLevenshteinDistance, GetLongestLeadingSubstring, PrettifyTitle
from fake_useragent import UserAgent
from markdown import markdown
from praw import Reddit
from praw.exceptions import InvalidURL
from praw.models import Submission
from prawcore.exceptions import Forbidden, InsufficientScope, NotFound
from time import sleep
import webbrowser

#
#
#
# The class definition.
#
#
#

class ExtractorReddit(Extractor):

    def __init__(self) -> None:

        ##
        #
        # The constructor.
        #
        ##

        super().__init__()

        self._downloadChapterSoupWhenExtracting = False

        self._userAgent = UserAgent().random

        if not ExtractorReddit._RefreshToken:
            self._redditInstance = Reddit(
                client_id = Configuration.RedditClientID,
                client_secret = None,
                redirect_uri = Configuration.RedditRedirectURI,
                user_agent = self._userAgent
            )

        else:
            self._redditInstance = Reddit(
                client_id = Configuration.RedditClientID,
                client_secret = None,
                refresh_token = ExtractorReddit._RefreshToken,
                user_agent = self._userAgent
            )

    def GetSupportedHostnames(self) -> List[str]:

        ##
        #
        # Returns a list of hostnames supposed to be supported by the extractor.
        #
        # @return A list of supported hostnames.
        #
        ##

        return [
            "reddit.com"
        ]

    def SupportsAuthentication(self) -> bool:

        ##
        #
        # Checks whether the extractor supports user authentication.
        #
        # @return **True** if the site *does* support authentication, **False** otherwise.
        #
        ##

        return True

    def Authenticate(self, interface: Interface) -> bool:

        ##
        #
        # Logs the user in, interactively.
        #
        # @param interface The user interface to be used.
        #
        # @return **True** if the user has been authenticated correctly, **False** otherwise.
        #
        ##

        if not ExtractorReddit._RefreshToken:

            scopes = [
                "identity",
                "read",
                "history",
            ]

            state = str(randint(0, 65000))
            authorizationURL = self._redditInstance.auth.url(scopes, state, "permanent")

            interface.Comment(
                f'You authorization URL is: "{authorizationURL}". {Configuration.ApplicationName}'
                " has just attempted to open it in your web browser - if it has failed, then please"
                " open it manually and confirm granting access to"
                f" {Configuration.ApplicationName}."
            )

            webbrowser.open(authorizationURL, new = 2)
            interface.GrabUserAttention()

            # Receive the connection.

            server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            server.bind(("localhost", 8080))
            server.listen(1)

            client = server.accept()[0]
            server.close()

            # Process received data.

            data = client.recv(1024).decode("utf-8")
            responseTemplate = ReadTextFile(GetPackageDirectory() / "Resources" / "PrettyHTMLText.html")
            responseTemplate = responseTemplate.replace("@@@ApplicationName@@@", Configuration.ApplicationName)
            responseTemplate = responseTemplate.replace("@@@ApplicationVersion@@@", Configuration.ApplicationVersion)

            if (" " not in data) or ("?" not in data) or ("&" not in data):
                SendMessage(client, responseTemplate.replace("@@@Content@@@", "Error: invalid response."))
                return self.AuthenticationResult.FAILURE

            parameterTokens = data.split(" ", 2)[1].split("?", 1)[1].split("&")
            parameters = {
                key: value
                for (key, value) in [token.split("=") for token in parameterTokens]
            }

            if parameters["state"] != state:
                SendMessage(client, responseTemplate.replace("@@@Content@@@", "Error: invalid state in response."))
                return self.AuthenticationResult.FAILURE

            elif "error" in parameters:
                SendMessage(client, responseTemplate.replace("@@@Content@@@", f'Error: {parameters["error"]}.'))
                return self.AuthenticationResult.FAILURE

            else:
                SendMessage(client, responseTemplate.replace("@@@Content@@@", "Everything went well. 😉"))

            # Authorize.

            ExtractorReddit._RefreshToken = self._redditInstance.auth.authorize(parameters["code"])

        # Re-create the reddit instance and notify the user.

        self._redditInstance = Reddit(
            client_id = Configuration.RedditClientID,
            client_secret = None,
            refresh_token = ExtractorReddit._RefreshToken,
            user_agent = self._userAgent
        )

        print(f'# You are authenticated as "{self._redditInstance.user.me()}".')

        return self.AuthenticationResult.SUCCESS

    def ScanStory(self) -> bool:

        ##
        #
        # Scans the story: generates the list of chapter URLs and retrieves the
        # metadata.
        #
        # @return **False** when the scan fails, **True** when it doesn't fail.
        #
        ##

        try:

            if self.Story is None:
                logging.error("The extractor isn't initialized.")
                return False

            # Retrieve chapter URLs.

            try:

                submission = Submission(self._redditInstance, url = self.Story.Metadata.URL)
                subredditName = submission.subreddit.display_name

                storyTitleProper = GetTitleProper(submission.title)
                if not storyTitleProper:
                    logging.error("Failed to read story title.")
                    return False

            except Forbidden:

                logging.error("PRAW says: Forbidden.")
                return False

            except InvalidURL:

                logging.error("PRAW says: InvalidURL.")
                return False

            except NotFound:

                logging.error("PRAW says: Forbidden.")
                return False

            if submission.author:

                try:

                    for nextSubmission in submission.author.submissions.new():

                        if subredditName != nextSubmission.subreddit.display_name:
                            continue

                        titleProper = GetTitleProper(nextSubmission.title)
                        if not titleProper:
                            continue

                        distance = GetLevenshteinDistance(storyTitleProper, titleProper)
                        if distance > 5:
                            continue

                        self._chapterURLs.append(nextSubmission.url)

                except (NotFound, InvalidURL, Forbidden):

                    pass

                self._chapterURLs.reverse()

            if not self._chapterURLs:
                self._chapterURLs = [submission.url]

            # Extract metadata.

            firstSubmission = Submission(self._redditInstance, url = self._chapterURLs[0])
            lastSubmission = Submission(self._redditInstance, url = self._chapterURLs[-1])

            storyTitleProper = PrettifyTitle(firstSubmission.title, removeContext = True)

            self.Story.Metadata.Title = storyTitleProper
            self.Story.Metadata.Author = (
                firstSubmission.author.name
                if firstSubmission.author else
                "[deleted]"
            )
            self.Story.Metadata.DatePublished = GetDateFromTimestamp(int(firstSubmission.created_utc))
            self.Story.Metadata.DateUpdated = GetDateFromTimestamp(int(lastSubmission.created_utc))
            self.Story.Metadata.ChapterCount = len(self._chapterURLs)
            self.Story.Metadata.WordCount = 0
            self.Story.Metadata.Summary = "No summary."

            # Return.

            return True

        except Forbidden:

            logging.error("PRAW says: Forbidden.")
            return False

        except InsufficientScope:

            logging.error("PRAW says: InsufficientScope.")
            return False

        except InvalidURL:

            logging.error("PRAW says: InvalidURL.")
            return False

        except NotFound:

            logging.error("PRAW says: Forbidden.")
            return False

    def _InternallyExtractChapter(
        self,
        URL: str,
        soup: Optional[BeautifulSoup]
    ) -> Optional[Chapter]:

        ##
        #
        # Extracts specific chapter.
        #
        # @param URL  The URL of the page containing the chapter.
        # @param soup The tag soup of the page containing the chapter.
        #
        # @return **True** if the chapter is extracted correctly, **False** otherwise.
        #
        ##

        submission = Submission(self._redditInstance, url = URL)

        return Chapter(content = markdown(submission.selftext))

    _RefreshToken = None

#
#
#
# Functions.
#
#
#

def SendMessage(client: socket.socket, message: str) -> None:

    ##
    #
    # Sends a message to client, then closes the connection.
    #
    # @param client  The client.
    # @param message The message.
    #
    ##

    if (not client) or (not message):
        return

    client.send("HTTP/1.1 200 OK\r\n\r\n{}".format(message).encode("utf-8"))
    client.close()



####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Application.

from fiction_dl.Concepts.Metadata import Metadata
from fiction_dl.Concepts.Story import Story
import fiction_dl.Configuration as Configuration

# Standard packages.

from typing import Callable, List

# Non-standard packages.

from dreamy_utilities.Containers import RemoveDuplicates
from dreamy_utilities.Text import FillTemplate, GetCurrentDate, PrettifyNumber

#
#
#
# Classes.
#
#
#

##
#
# Represents a group (list) of stories.
#
##

class StoryPackage:

    def __init__(
        self,
        stories: List[Story]
    ) -> None:

        ##
        #
        # The constructor.
        #
        # @param stories Stories included in the package.
        #
        ##

        # Initialize member variables.

        self.Metadata = Metadata(prettifyTitle = False)

        self.Stories = stories
        self.Images = []

        # Generate/calculate some metadata. Also extract images.

        totalAuthors = []
        totalChapterCount = 0
        totalWordCount = 0

        for story in self.Stories:

            totalAuthors.append(story.Metadata.Author)
            totalChapterCount += story.Metadata.ChapterCount
            totalWordCount += story.Metadata.WordCount

            self.Images.extend(story.Images)

        totalAuthors = RemoveDuplicates(totalAuthors)

        # Initialize the metadata.

        self.Metadata.URL = "n/a"

        self.Metadata.Title = f"{Configuration.ApplicationName} Package ({GetCurrentDate()})"
        self.Metadata.Author = ", ".join(totalAuthors)
        self.Metadata.Summary = "n/a"

        self.Metadata.DatePublished = stories[0].Metadata.DateExtracted
        self.Metadata.DateUpdated = stories[0].Metadata.DateExtracted
        self.Metadata.DateExtracted = stories[0].Metadata.DateExtracted

        self.Metadata.ChapterCount = totalChapterCount
        self.Metadata.WordCount = totalWordCount

    def FillTemplate(self, template: str, escapeHTMLEntities: bool = False) -> str:

        ##
        #
        # Replaces placeholders with appropriate values.
        #
        # @param template The code of the template to be filled.
        #
        # @return Filled template.
        #
        ##

        template = FillTemplate(Configuration, template)
        template = FillTemplate(self.Metadata.GetPrettified(escapeHTMLEntities), template)
        template = template.replace("@@@StoryCount@@@", PrettifyNumber(len(self.Stories)))

        return template

    def Join(
        self,
        prefixer: Callable[[int, str, str], str] = None,
        processor: Callable[[str], str] = None
    ) -> str:

        ##
        #
        # Returns the whole content of all the stories, with each chapter prefixed with the given
        # prefix and its content processed using given the given processor.
        #
        # @param prefixer  A function accepting (index, chapter title, story title) and returning a string. Optional.
        # @param processor A function used to process the content of each chapter. Takes a
        #                  string and returns a string.
        #
        # @return Processed story content.
        #
        ##

        joinedContent = ""

        for story in self.Stories:

            prettifiedTitle = story.Metadata.GetPrettified().Title

            for index, chapter in enumerate(story.Chapters, start = 1):

                prefix = prefixer(index, chapter.Title, prettifiedTitle) if prefixer else ""
                content = processor(chapter.Content) if processor else chapter.Content

                joinedContent += (prefix  + content)

        return joinedContent


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Application.

from fiction_dl.Concepts.Chapter import Chapter
from fiction_dl.Concepts.Extractor import Extractor
from fiction_dl.Concepts.Story import Story
from fiction_dl.Utilities.HTML import StripHTML

# Standard packages.

import logging
import re
import requests
from typing import List, Optional

# Non-standard packages.

from bs4 import BeautifulSoup
from dreamy_utilities.Text import Stringify
from dreamy_utilities.Web import GetHostname

#
#
#
# Classes.
#
#
#

##
#
# An extractor dedicated for FicWad.com.
#
##

class ExtractorFicWad(Extractor):

    def __init__(self) -> None:

        ##
        #
        # The constructor.
        #
        ##

        super().__init__()

    def GetSupportedHostnames(self) -> List[str]:

        ##
        #
        # Returns a list of hostnames supposed to be supported by the extractor.
        #
        # @return A list of supported hostnames.
        #
        ##

        return [
            "ficwad.com",
        ]

    def ScanChannel(self, URL: str) -> Optional[List[str]]:

        ##
        #
        # Scans the channel: generates the list of story URLs.
        #
        # @return **None** when the scan fails, a list of story URLs when it doesn't fail.
        #
        ##

        if (not URL) or (GetHostname(URL) not in self.GetSupportedHostnames()):
            return None

        if "/a/" not in URL:
            return None

        soup = self._webSession.GetSoup(URL)
        if not soup:
            logging.error(f"Couldn't download page: \"{URL}\".")
            return None

        storyURLs = []

        for element in soup.select(".storylist > li > h4 > a"):

            if not element.has_attr("href"):
                continue

            storyURLs.append(self._baseWorkURL + element["href"])

        return storyURLs

    def _InternallyScanStory(
        self,
        URL: str,
        soup: Optional[BeautifulSoup]
    ) -> bool:

        ##
        #
        # Scans the story: generates the list of chapter URLs and retrieves the
        # metadata.
        #
        # @param URL  The URL of the story.
        # @param soup The tag soup.
        #
        # @return **False** when the scan fails, **True** when it doesn't fail.
        #
        ##

        # Locate relevant page elements.

        titleElement = \
            soup.select_one(".storylist .blocked h4") \
            or \
            soup.select_one(".storylist h4 a")
        if not titleElement:
            logging.error("Title element not found.")
            return False

        authorElement = soup.select_one("span.author a")
        if not authorElement:
            logging.error("Author element not found.")
            return False

        summaryElement = soup.select_one("blockquote.summary")
        if not summaryElement:
            logging.error("Summary element not found.")
            return False

        metadataElement = soup.select_one("p.meta")
        if not metadataElement:
            logging.error("Metadata element not found.")
            return False

        # Extract metadata from the description.

        metadataText = metadataElement.get_text().strip()

        chapterCountMatch = re.search("Chapters:\s*(\d+)", metadataText)

        datePublishedMatch = re.search("Published:\s*([0-9-]+)", metadataText)
        if not datePublishedMatch:
            logging.error("Date published not found.")
            return False

        dateUpdatedMatch = re.search("Updated:\s*([0-9-]+)", metadataText)

        wordCountMatch = re.search("(\d+)\s*words", metadataText)
        if not wordCountMatch:
            logging.error("Word count not found.")
            return False

        # Set the metadata.

        self.Story.Metadata.Title = titleElement.get_text().strip()
        self.Story.Metadata.Author = authorElement.get_text().strip()

        self.Story.Metadata.DatePublished = datePublishedMatch.group(1)
        self.Story.Metadata.DateUpdated = \
            dateUpdatedMatch.group(1) if dateUpdatedMatch else self.Story.Metadata.DatePublished

        self.Story.Metadata.ChapterCount = \
            int(chapterCountMatch.group(1)) if chapterCountMatch else 1
        self.Story.Metadata.WordCount = int(wordCountMatch.group(1))

        self.Story.Metadata.Summary = StripHTML(summaryElement.get_text().strip())

        # Extract chapter URLs.

        for element in soup.select("#chapters .storylist li h4 a"):

            if not element.has_attr("href"):
                continue

            self._chapterURLs.append(self._baseWorkURL + element["href"])

        if not self._chapterURLs:
            self._chapterURLs = [self.Story.Metadata.URL]

        # Return.

        return True

    def _InternallyExtractChapter(
        self,
        URL: str,
        soup: Optional[BeautifulSoup]
    ) -> Optional[Chapter]:

        ##
        #
        # Extracts specific chapter.
        #
        # @param URL  The URL of the page containing the chapter.
        # @param soup The tag soup of the page containing the chapter.
        #
        # @return **True** if the chapter is extracted correctly, **False** otherwise.
        #
        ##

        # Locate relevant page elements.

        titleElement = \
            soup.select_one(".storylist .blocked h4") \
            or \
            soup.select_one(".storylist h4 a")
        if not titleElement:
            logging.error("Title element not found.")
            return False

        contentElement = soup.select_one("#storytext")
        if not contentElement:
            logging.error("Content element not found.")
            return False

        # Return.

        return Chapter(
            titleElement.get_text().strip(),
            Stringify(contentElement.encode_contents())
        )

    _baseWorkURL = "https://ficwad.com"


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Standard packages.

from io import BytesIO
import logging
from typing import Optional

# Non-standard packages.

import cv2
import numpy
import PIL.Image

#
#
#
# Classes.
#
#
#

##
#
# Represents an image.
#
##

class Image:

    def __init__(self, URL: str) -> None:

        ##
        #
        # The constructor.
        #
        # @param URL The image's URL.
        #
        ##

        self.URL = URL

        self.Data = None
        self.W = None
        self.H = None

    def CreateFromData(self, data: bytes, side: Optional[int] = None, quality: int = 75) -> bool:

        ##
        #
        # Creates an image from binary data. The image is then optionally scaled down to fit the maximum side length
        # provided as an argument.
        #
        # @param data    Encoded image data.
        # @param side    The maximum length of the longer side of the image. The image
        #                will be scaled down proportionally to fit.
        # @param quality The quality of the output image (1 - 100).
        #
        # @return **True** if the image has been created correctly, **False** otherwise.
        #
        ##

        if not data:
            return False

        try:

            processedImage =                                        \
                CreateImageFromDataUsingOpenCV(data, side, quality) \
                or                                                  \
                CreateImageFromDataUsingPIL(data, side, quality)

            if processedImage is None:
                logging.info(f"Failed to create image from data: \"{self.URL}\".")
                return False

        except:

            logging.info(f"An exception has occurred while creating image from data: \"{self.URL}\".")
            return False

        self.Data = processedImage[0]
        self.W = processedImage[1]
        self.H = processedImage[2]

        return True

    def __bool__(self) -> bool:

        ##
        #
        # The bool operator.
        #
        # @return **True** if the image exists (i.e. has data), **False** otherwise. An image that
        #         has its URL defined, but possesses no content doesn't exist (returns **False**).
        #
        ##

        return bool(self.Data)

#
#
#
# Functions.
#
#
#

def CreateImageFromDataUsingOpenCV(data: bytes, side: Optional[int] = None, quality: int = 75) -> bool:

    ##
    #
    # Creates an image from binary data. The image is then optionally scaled down to fit the maximum side length
    # provided as an argument.
    #
    # @param data    Encoded image data.
    # @param side    The maximum length of the longer side of the image. The image will be scaled down proportionally
    #                to fit.
    # @param quality The quality of the output image (1 - 100).
    #
    # @return A tuple consisting of encoded image data, image width and image height; **None** if something fails.
    #
    ##

    if not data:
        return None

    # Decode the image.

    image = cv2.imdecode(numpy.frombuffer(data, dtype = numpy.uint8), flags = -1)
    if (image is None) or (0 == image.size):
        return None

    height = image.shape[0]
    width = image.shape[1]
    if (not height) or (not width):
        return None

    # Scale down the image (if necessary).

    if side and (scale := side / max(width, height)) < 1:
        image = cv2.resize(
            image,
            (int(scale * width), int(scale * height)),
            interpolation = cv2.INTER_LANCZOS4
        )

    # Encode image data and store it.

    _, encodedImage = cv2.imencode(".jpg", image, [int(cv2.IMWRITE_JPEG_QUALITY), quality])
    convertedData = numpy.array(encodedImage).tobytes()

    if not convertedData:
        return None

    return (
        convertedData,
        image.shape[1],
        image.shape[0]
    )

def CreateImageFromDataUsingPIL(data: bytes, side: Optional[int] = None, quality: int = 75) -> bool:

    ##
    #
    # Creates an image from binary data. The image is then optionally scaled down to fit the maximum side length
    # provided as an argument.
    #
    # @param data    Encoded image data.
    # @param side    The maximum length of the longer side of the image. The image will be scaled down proportionally
    #                to fit.
    # @param quality The quality of the output image (1 - 100).
    #
    # @return A tuple consisting of encoded image data, image width and image height; **None** if something fails.
    #
    ##

    if not data:
        return None

    # Decode the image.

    image = PIL.Image.open(BytesIO(data)).convert("RGB")
    if not image:
        return None

    width = image.size[0]
    height = image.size[1]

    # Scale down the image (if necessary).

    if side and (scale := side / max(width, height)) < 1:
        image = image.resize(
            (int(scale * width), int(scale * height)),
            PIL.Image.ANTIALIAS
        )

    # Encode image data and store it.

    convertedData = BytesIO()
    image.save(
        convertedData,
        format = "JPEG",
        quality = quality,
        optimize = True
    )

    if not convertedData:
        return None

    return (
        convertedData.getvalue(),
        image.size[0],
        image.size[1]
    )


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Standard packages.

from pathlib import Path
from shutil import rmtree
from typing import Any

# Non-standard packages.

from bs4 import BeautifulSoup
from dreamy_utilities.Filesystem import GetUniqueFileName, ReadTextFile, WriteTextFile
from dreamy_utilities.Text import Bytify

#
#
#
# Classes.
#
#
#

##
#
# Represents the application's cache.
#
##

class Cache:

    def __init__(self, directoryPath: Path) -> None:

        ##
        #
        # The constructor.
        #
        # @param directoryPath The path of the cache's directory. It will be created if necessary.
        #                      Needs to be write-able.
        #
        ##

        self._directoryPath = directoryPath
        self._items = {}

        # Create the directory and attempt to read pre-existing cache.

        self._directoryPath.mkdir(parents = True, exist_ok = True)
        self._ReadIndexFromFile()

    def AddItem(self, owner: str, name: str, data: Any) -> None:

        ##
        #
        # Adds an item to the cache.
        #
        # @param owner The namespace.
        # @param name  The name of the item.
        # @param data  Data to be stored.
        #
        ##

        if (not owner) or (not name) or (not data):
            return

        if self.ContainsItem(owner, name):
            Path(self._items[owner][name]).unlink()

        filePath = self._directoryPath / GetUniqueFileName()
        with open(filePath, "wb") as file:
            file.write(Bytify(data))

        self._items.setdefault(owner, {})
        self._items[owner][name] = filePath

        self._SaveIndexToFile()

    def RetrieveItem(self, owner: str, name: str) -> bytes:

        ##
        #
        # Reads an item from the cache.
        #
        # @param owner The namespace.
        # @param item  The name of the item.
        #
        # @return The data associated with the item.
        #
        ##

        if (owner not in self._items) or (name not in self._items[owner]):
            return None

        with open(self._items[owner][name], "rb") as file:
            return file.read()

    def ContainsItem(self, owner: str, name: str) -> bool:

        ##
        #
        # Checks whether the cache contains given item.
        #
        # @param owner The namespace.
        # @param item  The name of the item.
        #
        # @return **True** if the item is present in the cache, **False** otherwise.
        #
        ##

        return (owner in self._items) and (name in self._items[owner])

    def Clear(self) -> None:

        ##
        #
        # Clears the cache.
        #
        ##

        self._items.clear()

        rmtree(self._directoryPath)
        self._directoryPath.mkdir(parents = True, exist_ok = True)

    def _ReadIndexFromFile(self) -> None:

        ##
        #
        # Reads the contents of the cache from the index file.
        #
        ##

        code = ReadTextFile(self._directoryPath / self._IndexFileName)
        if not code:
            return

        soup = BeautifulSoup(code, "xml")

        indexNode = soup.find("Index")
        if not indexNode:
            return

        for ownerNode in indexNode.find_all(recursive = False):

            ownerName = ownerNode.find("Name").get_text()
            self._items[ownerName] = {}

            for itemNode in ownerNode.find("Items").find_all(recursive = False):

                itemName = itemNode.find("Name").get_text()
                itemValue = itemNode.find("Value").get_text()

                self._items[ownerName][itemName] = itemValue

    def _SaveIndexToFile(self) -> None:

        ##
        #
        # Saves the index to the index file.
        #
        ##

        soup = BeautifulSoup(features = "xml")

        indexNode = soup.new_tag("Index")
        soup.append(indexNode)

        for ownerName, itemDictionary in self._items.items():

            ownerNode = soup.new_tag("Owner")
            indexNode.append(ownerNode)

            ownerNameNode = soup.new_tag("Name")
            ownerNameNode.string = ownerName
            ownerNode.append(ownerNameNode)

            ownerItemsNode = soup.new_tag("Items")
            ownerNode.append(ownerItemsNode)

            for name, value in itemDictionary.items():

                itemNode = soup.new_tag("Item")
                ownerItemsNode.append(itemNode)

                nameNode = soup.new_tag("Name")
                nameNode.string = name
                itemNode.append(nameNode)

                valueNode = soup.new_tag("Value")
                valueNode.string = str(value)
                itemNode.append(valueNode)

        WriteTextFile(self._directoryPath / self._IndexFileName, str(soup))

    _IndexFileName = "Index.xml"


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Application.

from fiction_dl.Concepts.Chapter import Chapter
from fiction_dl.Concepts.Extractor import Extractor

# Standard packages.

import logging
import re
from typing import List, Optional

# Non-standard packages.

from bs4 import BeautifulSoup
from dreamy_utilities.HTML import ReadElementText
from dreamy_utilities.Text import Stringify
from dreamy_utilities.Web import GetHostname, GetSiteURL

#
#
#
# Classes.
#
#
#

##
#
# An extractor dedicated for HarryPotterFanFiction.com.
#
##

class ExtractorHPFF(Extractor):

    def GetSupportedHostnames(self) -> List[str]:

        ##
        #
        # Returns a list of hostnames supposed to be supported by the extractor.
        #
        # @return A list of supported hostnames.
        #
        ##

        return [
            "harrypotterfanfiction.com",
        ]

    def ScanChannel(self, URL: str) -> Optional[List[str]]:

        ##
        #
        # Scans the channel: generates the list of story URLs.
        #
        # @return **None** when the scan fails, a list of story URLs when it doesn't fail.
        #
        ##

        if (not URL) or (GetHostname(URL) not in self.GetSupportedHostnames()):
            return None

        userIDMatch = re.search("uid=(\d+)", URL)
        if not userIDMatch:
            return None

        userID = userIDMatch.group(1)

        normalizedURL = f"{self.BASE_URL}/viewuser.php?uid={userID}"
        soup = self._webSession.GetSoup(normalizedURL)
        if not soup:
            logging.error(f"Couldn't download page: \"{normalizedURL}\".")
            return None

        storyURLs = []

        for element in soup.select("div#all-stories article.story-summary h3 a"):

            if not element.has_attr("href"):
                continue

            storyURLs.append(self.BASE_URL + element["href"])

        return storyURLs

    def _InternallyScanStory(
        self,
        URL: str,
        soup: Optional[BeautifulSoup]
    ) -> bool:

        ##
        #
        # Scans the story: generates the list of chapter URLs and retrieves the
        # metadata.
        #
        # @param URL  The URL of the story.
        # @param soup The tag soup.
        #
        # @return **False** when the scan fails, **True** when it doesn't fail.
        #
        ##

        # Extract the metadata.

        titleElement = soup.select_one("div.section__content > h2")
        if not titleElement:
            logging.error("Title element not found.")
            return False

        authorElement = soup.select_one("div.section__content > h2 > i > a")
        if not authorElement:
            logging.error("Author element not found.")
            return False

        rowElements = soup.select("article.section__inner > div.row")
        if len(rowElements) < 2:
            logging.error("Failed to retrieve the second row of metadata.")
            return False

        metadataRowElement = rowElements[1]
        metadataColumnElements = metadataRowElement.select("div.col")
        if len(metadataColumnElements) < 2:
            logging.error("Found not enough columns of metadata.")
            return False

        firstColumnEntriesElements = metadataColumnElements[0].select("div.entry")
        secondColumnEntriesElements = metadataColumnElements[1].select("div.entry")
        columnEntriesElements = firstColumnEntriesElements + secondColumnEntriesElements

        chapterCountElement = None
        wordCountElement = None
        datePublishedElement = None
        dateUpdatedElement = None

        for entryElement in columnEntriesElements:

            keyElement = entryElement.select_one("div.entry__key")
            if not keyElement:
                continue

            valueElement = entryElement.select_one("div.entry__value")
            if not valueElement:
                continue

            key = keyElement.get_text().strip()

            if "Chapters" == key:
                chapterCountElement = valueElement
            elif "Words" == key:
                wordCountElement = valueElement
            elif "First Published" == key:
                datePublishedElement = valueElement
            elif "Last Updated" == key:
                dateUpdatedElement = valueElement

        if (not chapterCountElement)  or \
           (not wordCountElement)     or \
           (not datePublishedElement) or \
           (not dateUpdatedElement):
            logging.error("Some important metadata not found.")
            return False

        # Process found elements.

        author = authorElement.get_text().strip()

        titleAuthorElement = titleElement.select_one("i")
        if titleAuthorElement:
            titleAuthorElement.decompose()

        # Set the metadata.

        self.Story.Metadata.Title = titleElement.get_text().strip()
        self.Story.Metadata.Author = author

        self.Story.Metadata.DatePublished = datePublishedElement.get_text().strip()[:10]
        self.Story.Metadata.DateUpdated = dateUpdatedElement.get_text().strip()[:10]

        self.Story.Metadata.ChapterCount = int(chapterCountElement.get_text().strip())
        self.Story.Metadata.WordCount = int(wordCountElement.get_text().strip().replace(",", ""))

        self.Story.Metadata.Summary = "No summary."

        # Retrieve chapter URLs.

        chapterRowElements = soup.select("table.table-chapters > tbody > tr")
        if not chapterRowElements:
            logging.error("Chapters not found.")
            return False

        baseURL = GetSiteURL(self.Story.Metadata.URL)

        for rowElement in chapterRowElements:

            anchorElement = rowElement.select_one("td > a.h4")
            if (not anchorElement) or not anchorElement.has_attr("href"):
                logging.warning("Chapter link present, yet can't be parsed.")

            self._chapterURLs.append(f"{baseURL}{anchorElement['href']}")

        # Return.

        self.Story.URL = "https://www.google.com/"

        return True

    def _InternallyExtractChapter(
        self,
        URL: str,
        soup: Optional[BeautifulSoup]
    ) -> Optional[Chapter]:

        ##
        #
        # Extracts specific chapter.
        #
        # @param URL  The URL of the page containing the chapter.
        # @param soup The tag soup of the page containing the chapter.
        #
        # @return **True** if the chapter is extracted correctly, **False** otherwise.
        #
        ##

        # Extract the title.

        title = None

        titleElement = soup.select_one("p.highlighted-image__title > a")
        if titleElement:
            title = titleElement.get_text().strip()

        # Extract the content.

        contentElement = soup.select_one("div.storytext-container")
        if not contentElement:
            logging.error("Could find the content element.")
            return None

        # Return.

        return Chapter(
            title = title,
            content = Stringify(contentElement.encode_contents())
        )

    def _GetNormalizedStoryURL(self, URL: str) -> str:

        ##
        #
        # Returns a normalized story URL, i.e. one that can be used for anything.
        #
        # @param URL Input URL (given by the user).
        #
        # @return Normalized URL.
        #
        ##

        return f"{URL}&showRestricted"

    BASE_URL = "https://harrypotterfanfiction.com"


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports
#
#
#

# Standard packages.

from pathlib import Path
import tempfile

#
#
#
# Globals.
#
#
#

ApplicationName = "fiction-dl"
ApplicationVersion = "1.8.2"
ApplicationShortDescription = (
    "A content downloader, capable of retrieving works of (fan)fiction from the web and saving them in a few common file"
    " formats."
)
ApplicationURL = "http://github.com/DreamCobbler/fiction-dl"

CreatorName = "Benedykt Synakiewicz"
CreatorContact = "dreamcobbler@outlook.com"

RedditClientID = "ScszEQn1cI7GgQ"
RedditRedirectURI = "http://localhost:8080"

CacheDirectoryPath = tempfile.gettempdir() / Path(f"{ApplicationName} Cache")
DebugDirectoryPath = Path(f"{ApplicationName} Debug Data")
OutputDirectoryPath = Path(f"{ApplicationName} Downloads")
SkippedURLsFilePath = Path(f"{ApplicationName} Skipped URLs.txt")

# A simple line-break.
LineBreak = 79 * "-"

# The length of a progress bar.
ProgressBarLength = 45

# The welcoming message, printed when the application starts.
WelcomingMessage = f"{ApplicationName} {ApplicationVersion}, by {CreatorName} ({ApplicationURL})"

# Maximum length of the longer side of an embedded image.
MaximumImageSideLength = 800

# The time application waits after downloading a chapter, in seconds.
PostChapterSleepTime = 1.0

# About repeated connection attempts.
MaximumConnectionAttemptCount = 10
ConnectionAttemptWait = 2.0

# The text that's supposed to be the in the beginning of a text source file.
TextSourceFileMagicText = "LOCAL TEXT STORY"

# The text that's supposed to serve as a chapter break in a text source file.
TextSourceFileChapterBreak = "CHAPTER BREAK LINE"



####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

from __future__ import annotations

# Application.

from fiction_dl.Processors.TypographyProcessor import TypographyProcessor

# Standard packages.

from copy import deepcopy
from typing import List, Optional

# Non-standard packages.

from dreamy_utilities.HTML import EscapeHTMLEntities, UnescapeHTMLEntities
from dreamy_utilities.Text import PrettifyDate, PrettifyNumber, PrettifyTitle, Truncate

#
#
#
# Classes.
#
#
#

##
#
# Represents a story's metadata.
#
##

class Metadata:

    def __init__(self, prettifyTitle: bool = True) -> None:

        ##
        #
        # The constructor.
        #
        # @param prettifyTitle Should the title be prettified when returning prettified metadata?
        #
        ##

        # Initialize member variables.

        self.URL = None

        self.Title = None
        self.Author = None
        self.Summary = None

        # All dates are ISO 8601 dates (YYYY-MM-DD).
        self.DatePublished = None
        self.DateUpdated = None
        self.DateExtracted = None

        self.ChapterCount = None
        self.WordCount = None

        # Save the options.

        self._prettifyTitle = prettifyTitle

    def AreValuesMissing(self) -> List[str]:

        ##
        #
        # Checks whether the object contains all the information it's meant to contain.
        #
        # @return A list of missing values.
        #
        ##

        missingValues = []

        if not self.URL:
            missingValues.append("URL")

        if not self.Title:
            missingValues.append("Title")

        if not self.Author:
            missingValues.append("Author")

        if not self.DatePublished:
            missingValues.append("DatePublished")

        if not self.DateUpdated:
            missingValues.append("DateUpdated")

        if not self.DateExtracted:
            missingValues.append("DateExtracted")

        if not self.ChapterCount:
            missingValues.append("ChapterCount")

        return missingValues

    def GetPrettified(self, escapeHTMLEntities: bool = False) -> Metadata:

        ##
        #
        # Returns prettified metadata (to be used when printing anything).
        #
        # @return Prettified metadata.
        #
        ##

        metadata = deepcopy(self)
        metadata.Process(summaryLength = None)

        metadata.DatePublished = PrettifyDate(self.DatePublished)
        metadata.DateUpdated = PrettifyDate(self.DateUpdated)
        metadata.DateExtracted = PrettifyDate(self.DateExtracted)

        metadata.ChapterCount = PrettifyNumber(self.ChapterCount)
        metadata.WordCount = PrettifyNumber(self.WordCount)

        if escapeHTMLEntities:

            metadata.URL = EscapeHTMLEntities(metadata.URL)
            metadata.Title = EscapeHTMLEntities(metadata.Title)
            metadata.Author = EscapeHTMLEntities(metadata.Author)
            metadata.Summary = EscapeHTMLEntities(metadata.Summary)

            metadata.DatePublished = EscapeHTMLEntities(metadata.DatePublished)
            metadata.DateUpdated = EscapeHTMLEntities(metadata.DateUpdated)
            metadata.DateExtracted = EscapeHTMLEntities(metadata.DateExtracted)

            metadata.ChapterCount = EscapeHTMLEntities(metadata.ChapterCount)
            metadata.WordCount = EscapeHTMLEntities(metadata.WordCount)

        return metadata

    def Process(self, summaryLength: Optional[int] = 250) -> None:

        ##
        #
        # Processes the metadata. To be used before generating output files.
        #
        ##

        prettifiedTitle = PrettifyTitle(self.Title, removeContext = True) if self._prettifyTitle else self.Title

        typographyProcessor = TypographyProcessor()

        self.Title = typographyProcessor.Process(prettifiedTitle).strip()
        self.Title = UnescapeHTMLEntities(self.Title)

        self.Summary = typographyProcessor.Process(self.Summary).strip()
        self.Summary = UnescapeHTMLEntities(self.Summary)

        if summaryLength:
            self.Summary = Truncate(self.Summary, summaryLength)

    def __bool__(self) -> bool:

        ##
        #
        # The bool operator.
        #
        # @return **True** if the metadata exists.
        #
        ##

        return bool(self.Title)


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# If the script is run by filename, add the current directory to PATH. This is necessary to make all
# the imports work (otherwise "fiction_dl." would lead nowhere).

if __package__ is None:

    try:

        from Utilities.Filesystem import GetPackageDirectory

    except ImportError:

        from fiction_dl.Utilities.Filesystem import GetPackageDirectory

    from dreamy_utilities.Filesystem import AddToPATH

    AddToPATH(GetPackageDirectory().parent)

# Application.

from fiction_dl.Core.Application import Application
from fiction_dl.Utilities.Filesystem import GetLibreOfficeExecutablePath
import fiction_dl.Configuration as Configuration

# Standard packages.

from argparse import ArgumentParser, Namespace
from http.client import RemoteDisconnected
import logging
from pathlib import Path
from requests.exceptions import ConnectionError
from sys import exit
from urllib3.exceptions import ProtocolError

#
#
#
# Functions.
#
#
#

def Main() -> None:

    ##
    #
    # The "Main" function of the application.
    #
    ##

    # Read command-line arguments.

    arguments = ReadCommandLineArguments()

    # Set up the logger.

    logging.basicConfig(level = logging.INFO if arguments.Verbose else logging.ERROR)

    # Create and launch the application.

    try:

        Application(
            arguments = arguments,
            cacheDirectoryPath = Configuration.CacheDirectoryPath
        ).Launch()

    except (ConnectionError, ProtocolError, RemoteDisconnected):

        print()
        print("# A connection error has occurred.")

        exit()

def ReadCommandLineArguments() -> Namespace:

    ##
    #
    # Reads the command-line arguments.
    #
    # @return The command-line arguments.
    #
    ##

    epilogMessage =                                                                                              \
        "You can use the application to auto-format a story saved in a text file.\r\n"                           \
        "\r\n"                                                                                                   \
        f"Put the text \"{Configuration.TextSourceFileMagicText}\" in the very first line of the file, then\r\n" \
        "the story's original URL in the next one, then the title in the third one,\r\n"                         \
        "then the author's name in the fourth, the one line summary in the fifth.\r\n"                           \
        "\r\n"                                                                                                   \
        "Then put the HTML-formatted story content in the line after that."                                      \
        "\r\n"                                                                                                   \
        f"You can put \"{Configuration.TextSourceFileChapterBreak}\" in a line in order to force a chapter break."

    argumentParser = ArgumentParser(
        description = Configuration.ApplicationShortDescription,
        epilog = epilogMessage
    )

    argumentParser.add_argument(
        "-a",
        dest = "Authenticate",
        action = "store_true",
        help = "authenticates the user to supported sites using interactive mode"
    )

    argumentParser.add_argument(
        "-c",
        dest = "ClearCache",
        action = "store_true",
        help = "clears the cache before launching the application"
    )

    argumentParser.add_argument(
        "-pack",
        dest = "Pack",
        action = "store_true",
        help = "packs all downloaded stories inside one file (of each type)"
    )

    argumentParser.add_argument(
        "-v",
        dest = "Verbose",
        action = "store_true",
        help = "enables the (more) verbose mode"
    )

    argumentParser.add_argument(
        "-f",
        dest = "Force",
        action = "store_true",
        help = "overwrites output files (in case they already exist)"
    )

    argumentParser.add_argument(
        "-d",
        dest = "Debug",
        action = "store_true",
        help = "enables debug mode (saves some data useful for debugging)"
    )

    argumentParser.add_argument(
        "-no-images",
        dest = "Images",
        action = "store_false",
        help = "disables downloading images found in story content"
    )

    argumentParser.add_argument(
        "-persistent-cache",
        dest = "PersistentCache",
        action = "store_true",
        help = "preserves the cache after the application quits"
    )

    argumentParser.add_argument(
        "-lo",
        dest = "LibreOffice",
        type = Path,
        default = GetLibreOfficeExecutablePath() or Path(),
        help = "path to the LibreOffice executable (soffice.exe)"
    )

    argumentParser.add_argument(
        "-o",
        dest = "Output",
        type = str,
        default = Configuration.OutputDirectoryPath,
        help = "output directory path"
    )

    argumentParser.add_argument(
        "Input",
        type = str,
        help = "a URL, or a path to a text file with one URL per line"
    )

    return argumentParser.parse_args()

#
#
#
# The start-up routine.
#
#
#

if "__main__" == __name__:

    Main()



####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Application.

from fiction_dl.Concepts.Chapter import Chapter
from fiction_dl.Concepts.Extractor import Extractor
from fiction_dl.Utilities.HTML import StripHTML

# Standard packages.

from datetime import datetime
import logging
from math import ceil
import re
from typing import List, Optional

# Non-standard packages.

from bs4 import BeautifulSoup
from dreamy_utilities.Text import Stringify
from dreamy_utilities.Web import GetHostname

#
#
#
# Classes.
#
#
#

##
#
# An extractor dedicated for hentai-foundry.com.
#
##

class ExtractorHentaiFoundry(Extractor):

    def __init__(self) -> None:

        ##
        #
        # The constructor.
        #
        ##

        super().__init__()

    def GetSupportedHostnames(self) -> List[str]:

        ##
        #
        # Returns a list of hostnames supposed to be supported by the extractor.
        #
        # @return A list of supported hostnames.
        #
        ##

        return [
            "hentai-foundry.com"
        ]

    def ScanChannel(self, URL: str) -> Optional[List[str]]:

        ##
        #
        # Scans the channel: generates the list of story URLs.
        #
        # @return **None** when the scan fails, a list of story URLs when it doesn't fail.
        #
        ##

        if (not URL) or (GetHostname(URL) not in self.GetSupportedHostnames()):
            return None

        usernameStoryIDMatch = re.search("/user/([a-zA-Z0-9_]+)/(\d+)", URL)
        if usernameStoryIDMatch:
            return None

        usernameMatch = re.search("/user/([a-zA-Z0-9_]+)", URL)
        if not usernameMatch:
            return None

        username = usernameMatch.group(1)
        normalizedURL = f"http://www.hentai-foundry.com/stories/user/{username}/"

        pageSoup = self._webSession.GetSoup(self._GetAdultView(normalizedURL))
        if not pageSoup:
            return None

        pageCountDescriptionElement = pageSoup.select_one(".galleryHeader > .summary")
        pageCountDescription = pageCountDescriptionElement.get_text().strip()

        pageCountDescriptionMatch = re.search(
            "Displaying (\d+)-(\d+) of (\d+) results",
            pageCountDescription
        )

        if not pageCountDescriptionMatch:
            logging.error("Failed to retrieve page count of the Stories tab.")
            return None

        storiesPerPage = int(pageCountDescriptionMatch.group(2))
        storiesInTotal = int(pageCountDescriptionMatch.group(3))

        if not storiesPerPage:
            return None

        pageCount = ceil(storiesInTotal / storiesPerPage)

        storyURLs = []
        for pageIndex in range(1, pageCount + 1):

            pageURL = self._GetAdultView(
                f"http://www.hentai-foundry.com/stories/user/{username}?page={pageIndex}"
            )

            pageSoup = self._webSession.GetSoup(pageURL)
            if not pageSoup:
                return None

            storyLinkElements = pageSoup.select(".items > .storyRow > .titlebar > a")

            for linkElement in storyLinkElements:

                if not linkElement.has_attr("href"):
                    continue

                storyURLs.append(self._baseURL + linkElement["href"])

        return storyURLs

    def _InternallyScanStory(
        self,
        URL: str,
        soup: Optional[BeautifulSoup]
    ) -> bool:

        ##
        #
        # Scans the story: generates the list of chapter URLs and retrieves the
        # metadata.
        #
        # @param URL  The URL of the story.
        # @param soup The tag soup.
        #
        # @return **False** when the scan fails, **True** when it doesn't fail.
        #
        ##

        # Locate metadata.

        titleElement = soup.select_one(".titlebar a")
        if not titleElement:
            logging.error("Title element not found.")
            return False

        authorElement = soup.select_one(".storyInfo > .col1 > a")
        if not authorElement:
            logging.error("Author element not found.")
            return False

        datesElements = soup.select(".storyInfo > .col2 > .indent")
        if (not datesElements) or (len(datesElements) < 2):
            logging.error("Dates elements not found.")
            return False

        datePublishedElement = datesElements[0]
        dateUpdatedElement = datesElements[1]

        summaryElement = soup.select_one(".storyDescript")
        if not summaryElement:
            logging.error("Summary element not found.")
            return False

        chapterCountWordCountElement = soup.select_one(".storyInfo > .col3")
        if not chapterCountWordCountElement:
            logging.error("Chapter/word count elements not found.")
            return False

        # Extract and save metadata.

        self.Story.Metadata.Title = titleElement.get_text().strip()
        self.Story.Metadata.Author = authorElement.get_text().strip()

        rawDatePublished = datePublishedElement.get_text().strip()
        rawDateUpdated = dateUpdatedElement.get_text().strip()

        self.Story.Metadata.DatePublished = self._ReformatDate(rawDatePublished)
        self.Story.Metadata.DateUpdated = self._ReformatDate(rawDateUpdated)

        chapterCountWordCountDescription = StripHTML(
            chapterCountWordCountElement.get_text().strip()
        )
        chapterCountMatch = re.search("Chapters:\s+(\d+)", chapterCountWordCountDescription)
        if not chapterCountMatch:
            logging.error("Chapter count not found.")
            return False

        wordCountMatch = re.search("Words:\s+([0-9,]+)", chapterCountWordCountDescription)
        if not wordCountMatch:
            logging.error("Word count not found.")
            return False

        self.Story.Metadata.ChapterCount = int(chapterCountMatch.group(1))
        self.Story.Metadata.WordCount = self._ReadWordCount(wordCountMatch.group(1))

        self.Story.Metadata.Summary = StripHTML(summaryElement.get_text().strip())

        # Retrieve chapter URLs.

        chapterLinkElements = soup.select(".boxbody > p > a")
        if not chapterLinkElements:
            logging.error("No chapter links found.")
            return False

        for linkElement in chapterLinkElements:

            if not linkElement.has_attr("href"):
                continue

            self._chapterURLs.append(self._baseURL + linkElement["href"])

        # Return.

        return True

    def _InternallyExtractChapter(
        self,
        URL: str,
        soup: Optional[BeautifulSoup]
    ) -> Optional[Chapter]:

        ##
        #
        # Extracts specific chapter.
        #
        # @param URL  The URL of the page containing the chapter.
        # @param soup The tag soup of the page containing the chapter.
        #
        # @return **True** if the chapter is extracted correctly, **False** otherwise.
        #
        ##

        # Read the title.

        chapterTitle = None

        if (titleElement := soup.select_one("#viewChapter > .boxheader")):

            chapterTitle = titleElement.get_text().strip()

        # Read the content.

        storyTextElement = soup.select_one("#viewChapter > .boxbody")
        if not storyTextElement:
            logging.error("Story text element not found.")
            return None

        return Chapter(
            title = chapterTitle,
            content = Stringify(storyTextElement.encode_contents())
        )

    def _GetNormalizedStoryURL(self, URL: str) -> str:

        ##
        #
        # Returns a normalized story URL, i.e. one that can be used for anything.
        #
        # @param URL Input URL (given by the user).
        #
        # @return Normalized URL.
        #
        ##

        return self._GetAdultView(URL)

    @staticmethod
    def _GetAdultView(URL: str) -> Optional[str]:

        ##
        #
        # Returns a URL with option allowing to view adult content set.
        #
        # @param URL Input URL.
        #
        # @return Adult content proofed URL.
        #
        ##

        if not URL:
            return None

        return URL + "?enterAgree=1&size=0"

    @staticmethod
    def _ReformatDate(date: str) -> Optional[str]:

        if not date:
            return None

        try:

            date = datetime.strptime(date, "%B %d, %Y").strftime("%Y-%m-%d")

            return date

        except ValueError:

            return GetCurrentDate()

    @staticmethod
    def _ReadWordCount(wordsDescription: str) -> Optional[int]:

        ##
        #
        # Reads word count value from HF's "words" description.
        #
        # @param wordsDescription Input words description.
        #
        # @return Word count.
        #
        ##

        if not wordsDescription:
            return None

        wordCount = wordsDescription.strip().replace(",", "")

        try:

            wordCount = int(wordCount)

        except ValueError:

            logging.error("Failed to convert word count.")

            return None

        return wordCount

    _baseURL = "http://www.hentai-foundry.com"


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Application.

from fiction_dl.Concepts.Image import Image
from fiction_dl.Concepts.Metadata import Metadata
from fiction_dl.Utilities.HTML import StripHTML
import fiction_dl.Configuration as Configuration

# Standard packages.

from typing import Callable

# Non-standard packages.

from dreamy_utilities.Text import FillTemplate, GetCurrentDate

#
#
#
# Classes.
#
#
#

##
#
# Represents a story.
#
##

class Story:

    def __init__(self, URL: str = "") -> None:

        ##
        #
        # The constructor.
        #
        # @param URL The URL leading to the story.
        #
        ##

        self.Metadata = Metadata()

        self.Metadata.URL = URL
        self.Metadata.DatePublished = GetCurrentDate()
        self.Metadata.DateUpdated = self.Metadata.DatePublished
        self.Metadata.DateExtracted = self.Metadata.DatePublished

        self.Chapters = []
        self.Images = []

    def FillTemplate(self, template: str, escapeHTMLEntities: bool = False) -> str:

        ##
        #
        # Replaces placeholders with appropriate values.
        #
        # @param template The code of the template to be filled.
        #
        # @return Filled template.
        #
        ##

        template = FillTemplate(Configuration, template)
        template = FillTemplate(self.Metadata.GetPrettified(escapeHTMLEntities), template)

        return template

    def CalculateWordCount(self) -> int:

        ##
        #
        # Calculates the total word count of the story. **Calculates** it, not just returns the
        # .WordCount variable. Might take a moment for longer stories.
        #
        # @return The word count.
        #
        ##

        if not self.Chapters:
            return 0

        wordCount = 0

        for chapter in self.Chapters:
            wordCount += len(StripHTML(chapter.Content).split())

        return wordCount

    def Join(
        self,
        prefixer: Callable[[int, str, str], str] = None,
        processor: Callable[[str], str] = None
    ) -> str:

        ##
        #
        # Returns the whole content of the story, with each chapter prefixed with the given prefix
        # and its content processed using given the given processor.
        #
        # @param prefixer  A function accepting (index, chapter title, story title) and returning a string. Optional.
        # @param processor A function used to process the content of each chapter. Takes a
        #                  string and returns a string.
        #
        # @return Processed story content.
        #
        ##

        joinedContent = ""
        prettifiedTitle = self.Metadata.GetPrettified().Title

        for index, chapter in enumerate(self.Chapters, start = 1):

            prefix = prefixer(index, chapter.Title, prettifiedTitle) if prefixer else ""
            content = processor(chapter.Content) if processor else chapter.Content

            joinedContent += (prefix  + content)

        return joinedContent

    def Process(self) -> None:

        ##
        #
        # Processes the story metadata. To be used before generating output files.
        #
        ##

        self.Metadata.Process()

        for chapter in self.Chapters:
            chapter.Process()

    def __bool__(self) -> bool:

        ##
        #
        # The bool operator.
        #
        # @return **True** if the story has any content.
        #
        ##

        return bool(self.Metadata) and bool(self.Chapters)


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Application.

from fiction_dl.Concepts.Formatter import Formatter
from fiction_dl.Concepts.Story import Story
from fiction_dl.Concepts.StoryPackage import StoryPackage
from fiction_dl.Utilities.Filesystem import GetPackageDirectory
from fiction_dl.Utilities.HTML import StripEmptyTags

# Standard packages.

import html
from pathlib import Path
import re
from typing import List, Union
from zipfile import ZipFile, is_zipfile, ZIP_DEFLATED

# Non-standard packages.

from bs4 import BeautifulSoup
from dreamy_utilities.Filesystem import ReadTextFile
from dreamy_utilities.Mathematics import GetDimensionsToFit
from dreamy_utilities.Text import Stringify

#
#
#
# Classes.
#
#
#

##
#
# The ODT formatter.
#
##

class FormatterODT(Formatter):

    def __init__(self, embedImages: bool = True, combinedVersion: bool = False) -> None:

        ##
        #
        # The constructor.
        #
        # @param embedImages     Embed images in the output file.
        # @param combinedVersion Use the template designed for multiple stories.
        #
        ##

        super().__init__(embedImages)

        # Intialize member variables.

        templateFileName =                        \
            "Templates/FormatterODT/Template.odt" \
            if not combinedVersion else           \
            "Templates/FormatterODT/Template (Combined).odt"

        self._templateFilePath = GetPackageDirectory() / templateFileName

        self._manifestDocument = ""
        self._contentDocument = ""
        self._metadataDocument = ""
        self._stylesDocument = ""

        # Load the template.

        with ZipFile(self._templateFilePath, "r") as archive:

            self._manifestDocument = Stringify(archive.read("META-INF/manifest.xml"))
            self._contentDocument = Stringify(archive.read("content.xml"))
            self._metadataDocument = Stringify(archive.read("meta.xml"))
            self._stylesDocument = Stringify(archive.read("styles.xml"))

        # Modify the styles.

        EOF = self._stylesDocument.find("</office:styles>")

        styles = ReadTextFile(GetPackageDirectory() / "Templates/FormatterODT/Styles.xml")
        self._stylesDocument = self._stylesDocument[:EOF] + styles + self._stylesDocument[EOF:]

    def FormatAndSave(self, story: Union[Story, StoryPackage], filePath: Path) -> bool:

        ##
        #
        # Formats the story (or the story package) and saves it to the output file.
        #
        # @param story    The story/story package to be formatted and saved.
        # @param filePath The path to the output file.
        #
        # @return **True** if the output file was generated and saved without problems, **False**
        #         otherwise.
        #
        ##

        # Retrieve story content and translate it to ODT-compatible XML.

        def ChapterTitler(index: int, chapterTitle: str) -> str:
            return f"Chapter {index}" + (f": {chapterTitle}" if chapterTitle else "")

        def PackagePrefixer(index: int, chapterTitle: str, storyTitle: str) -> str:
            return f"<h1>{storyTitle} — {ChapterTitler(index, chapterTitle)}</h1>"

        def StoryPrefixer(index: int, chapterTitle: str, storyTitle: str) -> str:
            return f"<h1>{ChapterTitler(index, chapterTitle)}</h1>"

        prefixer = PackagePrefixer if isinstance(story, StoryPackage) else StoryPrefixer

        content = self._TranslateHTMLtoODT(story.Join(prefixer), story)
        content = StripEmptyTags(
            content,
            validEmptyTags = ["draw:frame", "draw:image"],
            validEmptyTagAttributes = {"text:style-name": "Horizontal_20_Line"}
        )

        # Prepare the files.

        manifestDocument = self._manifestDocument
        contentDocument = self._contentDocument
        metadataDocument = self._metadataDocument
        stylesDocument = self._stylesDocument

        # Modify the content.

        metadata = story.Metadata.GetPrettified(escapeHTMLEntities = True)

        contentDocument = story.FillTemplate(contentDocument, escapeHTMLEntities = True)
        contentDocument = contentDocument.replace("http://link.link/", metadata.URL)

        EOF = contentDocument.find("</office:text>")
        contentDocument = contentDocument[:EOF] + content + contentDocument[EOF:]

        # Modify the metadata.

        EOF = "</office:meta>"

        metadataDocument = self._SetTagContent(
            metadataDocument,
            "dc:title",
            html.escape(metadata.Title),
            EOF
        )

        metadataDocument = self._SetTagContent(
            metadataDocument,
            "meta:initial-creator",
            html.escape(metadata.Author),
            EOF
        )

        metadataDocument = self._SetTagContent(
            metadataDocument,
            "dc:creator",
            html.escape(metadata.Author),
            EOF
        )

        # Modify the styles.

        stylesDocument = story.FillTemplate(stylesDocument, escapeHTMLEntities = True)

        # Modify the manifest.

        if self._embedImages:

            for index, image in enumerate(story.Images):

                if not image.Data:
                    continue

                EOF = manifestDocument.find("</manifest:manifest>")
                manifestDocument = manifestDocument[:EOF] + \
                    '<manifest:file-entry manifest:full-path="Pictures/{}.jpeg"' \
                    ' manifest:media-type="image/jpeg"/>'.format(index) + \
                    manifestDocument[EOF:]

        # Save the output file.

        ReplacedFilesNames = [
            "META-INF/manifest.xml",
            "content.xml",
            "meta.xml",
            "styles.xml"
        ]

        with ZipFile(filePath, mode = "a") as outputArchive:

            with ZipFile(self._templateFilePath, "r") as archive:

                for item in [x for x in archive.infolist() if x.filename not in ReplacedFilesNames]:
                    outputArchive.writestr(item, archive.read(item.filename))

            outputArchive.writestr("META-INF/manifest.xml", manifestDocument)
            outputArchive.writestr("content.xml", contentDocument)
            outputArchive.writestr("meta.xml", metadataDocument)
            outputArchive.writestr("styles.xml", stylesDocument)

            if self._embedImages:

                for index, image in enumerate(story.Images):

                    if not image:
                        continue

                    outputArchive.writestr(f"Pictures/{index}.jpeg", image.Data)

        # Return.

        return True

    @staticmethod
    def _SetTagContent(
        document: str,
        tagName: str,
        newValue: str,
        EOF: str
    ) -> str:

        ##
        #
        # Replaces XML tag contents.
        #
        # @param document XML code.
        # @param tagName Tag name.
        # @param newValue Desired tag value.
        # @param EOF A string designating end-of-file.
        #
        # @return Modified code.
        #
        ##

        openingTag = f"<{tagName}>"
        closingTag = f"</{tagName}>"
        tagWithValue = f"{openingTag}{newValue}{closingTag}"

        if -1 != (tagPosition := document.find(openingTag)):

            document = re.sub(
                f"{openingTag}(.*){closingTag}",
                tagWithValue,
                document
            )

        elif -1 != (EOF := document.find(EOF)):

            document = document[:EOF] + tagWithValue + document[EOF:]

        # Return.

        return document

    def _TranslateHTMLtoODT(self, code: str, story: Union[Story, StoryPackage]) -> str:

        ##
        #
        # Translates HTML code to ODT code.
        #
        # @param code HTML code.
        #
        # @return ODT code.
        #
        ##

        # Process the content.

        soup = BeautifulSoup(code, features = "html.parser")

        imageIndex = 0

        for tag in soup.find_all(True, recursive = True):

            if "h1" == tag.name:

                tag.name = "text:h"
                tag["text:style-name"] = "Imported_Chapter"
                tag["text:outline-level"] = "1"

            elif "hr" == tag.name:

                tag.name = "text:p"
                tag["text:style-name"] = "Horizontal_20_Line"

            elif "p" == tag.name:

                tag.name = "text:p"
                tag["text:style-name"] = "Text_20_body"

            elif tag.name in ["b", "strong"]:

                tag.name = "text:span"
                tag["text:style-name"] = "Imported_Bold"

            elif tag.name in ["i", "em"]:

                tag.name = "text:span"
                tag["text:style-name"] = "Imported_Emphasis"

            elif "u" == tag.name:

                tag.name = "text:span"
                tag["text:style-name"] = "Imported_Underline"

            elif "a" == tag.name:

                if tag.has_attr("href"):

                    tag.name = "text:a"
                    tag["xlink:type"] = "simple"
                    tag["xlink:href"] = tag["href"]
                    tag["text:style-name"] = "Internet_20_link"
                    tag["text:visited-style-name"] = "Visited_20_Internet_20_Link"

                    del tag["href"]

                else:

                    tag.replaceWith("")

            elif ("img" == tag.name):

                if (not self._embedImages) or (imageIndex >= len(story.Images)):
                    tag.replaceWith("")
                    continue

                image = story.Images[imageIndex]
                if not image:
                    tag.replaceWith("")
                    continue

                maximumWidth = 17 # cm
                maximumHeight = 17 # cm
                width, height = GetDimensionsToFit(
                    (image.W, image.H),
                    (maximumWidth, maximumHeight)
                )

                drawFrameTag = soup.new_tag("draw:frame")
                drawFrameTag["draw:style-name"] = "Imported_Image"
                drawFrameTag["draw:name"] = f"Image{imageIndex}"
                drawFrameTag["text:anchor-type"] = "char"
                drawFrameTag["svg:width"] = "{:.3f}cm".format(width)
                drawFrameTag["svg:height"] = "{:.3f}cm".format(height)
                drawFrameTag["draw:z-index"] = "1"

                drawImageTag = soup.new_tag("draw:image")
                drawImageTag["xlink:href"] = f"Pictures/{imageIndex}.jpeg"
                drawImageTag["xlink:type"] = "simple"
                drawImageTag["xlink:show"] = "embed"
                drawImageTag["xlink:actuate"] = "onLoad"
                drawImageTag["loext:mime-type"] = "image/jpeg"
                drawFrameTag.append(drawImageTag)

                tag.replace_with(drawFrameTag)

                imageIndex += 1

        return str(soup)


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Application.

from fiction_dl.Concepts.Chapter import Chapter
from fiction_dl.Concepts.Extractor import Extractor
from fiction_dl.Utilities.HTML import MakeURLAbsolute

# Standard packages.

import logging
import re
import requests
from typing import List, Optional
from urllib.parse import urlparse

# Non-standard packages.

from bs4 import BeautifulSoup
from dreamy_utilities.Text import Stringify
from dreamy_utilities.Web import GetHostname, GetSiteURL

#
#
#
# Classes.
#
#
#

##
#
# The extractor designed for Adult-Fanfiction.org.
#
##

class ExtractorAdultFanfiction(Extractor):

    def __init__(self) -> None:

        ##
        #
        # The constructor.
        #
        ##

        super().__init__()

    def GetSupportedHostnames(self) -> List[str]:

        ##
        #
        # Returns a list of hostnames supposed to be supported by the extractor.
        #
        # @return A list of supported hostnames.
        #
        ##

        return [
            "adult-fanfiction.org"
        ]

    def ScanChannel(self, URL: str) -> Optional[List[str]]:

        ##
        #
        # Scans the channel: generates the list of story URLs.
        #
        # @return **None** when the scan fails, a list of story URLs when it doesn't fail.
        #
        ##

        if (not URL) or (GetHostname(URL) not in self.GetSupportedHostnames()):
            return None

        userIDMatch = re.search("profile\.php\?no=(\d+)", URL)
        if not userIDMatch:
            return None

        userID = userIDMatch.group(1)

        normalizedURL = f"http://members.adult-fanfiction.org/profile.php?no={userID}&view=story"
        soup = self._webSession.GetSoup(normalizedURL)
        if not soup:
            logging.error(f"Couldn't download page: \"{normalizedURL}\".")
            return None

        storyElements = self._FindAllStoriesByUserElements(userID)
        storyURLs = []

        for element in storyElements:

            anchorElement = element.select_one("a")
            if not anchorElement.has_attr("href"):
                continue

            storyURLs.append(anchorElement["href"])

        return storyURLs

    def _InternallyScanStory(
        self,
        URL: str,
        soup: Optional[BeautifulSoup]
    ) -> bool:

        ##
        #
        # Scans the story: generates the list of chapter URLs and retrieves the
        # metadata.
        #
        # @param URL  The URL of the story.
        # @param soup The tag soup.
        #
        # @return **False** when the scan fails, **True** when it doesn't fail.
        #
        ##

        # Extract chapter URLs.

        for element in soup.select("div.dropdown-content > a"):
            self._chapterURLs.append(
                MakeURLAbsolute(element["href"],
                GetSiteURL(self._GetNormalizedStoryURL(self.Story.Metadata.URL)))
            )

        # Find the author's profile.

        userProfileAnchorElement = soup.select_one("#contentdata tr > td > b > i > a")
        if not userProfileAnchorElement:
            logging.error("Author profile link element not found.")
            return False

        userProfileBaseURL = userProfileAnchorElement["href"]
        zoneName = urlparse(self.Story.Metadata.URL).hostname.split(".")[0]

        authorProfileURL = f"{userProfileBaseURL}&view=story&zone={zoneName}"
        authorProfileSoup = self._webSession.GetSoup(authorProfileURL)
        if not soup:
            logging.error(f'Failed to download page: "{authorProfileURL}".')
            return False

        authorElement = authorProfileSoup.select_one("div#contentdata > h2")
        if not authorElement:
            logging.error("Author element not found.")
            return False

        # Find the story's entry in the author's profile.

        matchingStoryLinkElement = None

        userProfileURL = userProfileAnchorElement["href"]
        userIDMatch = re.search("profile\.php\?no=(\d+)", userProfileURL)
        if not userIDMatch:
            return None

        userID = userIDMatch.group(1)
        storyID = self._GetIDFromURL(URL)

        allStoryElements = self._FindAllStoriesByUserElements(userID)
        for element in allStoryElements:

            anchorElement = element.select_one("a")

            if self._GetIDFromURL(anchorElement["href"]) == storyID:

                matchingStoryLinkElement = anchorElement.parent
                break

        if not matchingStoryLinkElement:
            logging.error("Story description in the author's profile page not found.")
            return False

        matchingStoryLinkElement.select_one("a").decompose()
        storyDescriptionContent = matchingStoryLinkElement.get_text().strip()

        locatedTextPosition = storyDescriptionContent.find("Located : ")
        if -1 == locatedTextPosition:
            logging.error("Story description doesn't conform to expected format.")
            return False

        summary = storyDescriptionContent[:locatedTextPosition].strip()

        publishedMatch = re.search("Posted \: (\d\d\d\d\-\d\d\-\d\d)", storyDescriptionContent)
        if not publishedMatch:
            logging.error("Couldn't find date published in story description.")
            return False

        updatedMatch = re.search("Edited \: (\d\d\d\d\-\d\d\-\d\d)", storyDescriptionContent)
        if not updatedMatch:
            logging.error("Couldn't find date updated in story description.")
            return False

        # Set the metadata.

        self.Story.Metadata.Title = soup.select_one("h2 > a").get_text().strip()
        self.Story.Metadata.Author = authorElement.get_text().strip()

        self.Story.Metadata.DatePublished = publishedMatch.group(1).strip()
        self.Story.Metadata.DateUpdated = updatedMatch.group(1).strip()

        self.Story.Metadata.ChapterCount = len(self._chapterURLs)
        self.Story.Metadata.WordCount = 0

        self.Story.Metadata.Summary = summary

        # Return.

        return True

    def _InternallyExtractChapter(
        self,
        URL: str,
        soup: Optional[BeautifulSoup]
    ) -> Optional[Chapter]:

        ##
        #
        # Extracts specific chapter.
        #
        # @param URL  The URL of the page containing the chapter.
        # @param soup The tag soup of the page containing the chapter.
        #
        # @return **True** if the chapter is extracted correctly, **False** otherwise.
        #
        ##

        rowElements = soup.select("div#contentdata > table > tr")
        if (not rowElements) or len(rowElements) < 3:
            logging.error("Chapter page doesn't conform to expected format.")

        return Chapter(
            title = None,
            content = Stringify(rowElements[2].encode_contents())
        )

    def _GetNormalizedStoryURL(self, URL: str) -> str:

        ##
        #
        # Returns a normalized story URL, i.e. one that can be used for anything.
        #
        # @param URL Input URL (given by the user).
        #
        # @return Normalized URL.
        #
        ##

        return re.sub("\&chapter\=(\d)+", "", URL)

    def _FindAllStoriesByUserElements(self, userID):

        normalizedURL = f"http://members.adult-fanfiction.org/profile.php?no={userID}&view=story"
        soup = self._webSession.GetSoup(normalizedURL)
        if not soup:
            logging.error(f"Couldn't download page: \"{normalizedURL}\".")
            return None

        zoneNames = []

        for element in soup.select("div#contentdata > div.alistnav a"):

            if not element.has_attr("href"):
                continue

            zoneNameMatch = re.search(
                "zone=([a-zA-Z0-9]+)",
                element["href"]
            )

            if not zoneNameMatch:
                continue

            zoneNames.append(zoneNameMatch.group(1))

        storyElements = []

        for zoneName in zoneNames:

            normalizedURL = f"http://members.adult-fanfiction.org/profile.php?no={userID}&view=story&zone={zoneName}"
            soup = self._webSession.GetSoup(normalizedURL)
            if not soup:
                logging.error(f"Couldn't download page: \"{normalizedURL}\".")
                return None

            lastPageIndex = 1

            if (paginationElements := soup.select("div#contentdata > div.pagination > ul > li > a")):

                paginationElement = paginationElements[-1]

                lastPageIndexMatch = re.search("page=(\d+)", paginationElement["href"])
                lastPageIndex = int(lastPageIndexMatch.group(1))

            for pageIndex in range(1, lastPageIndex + 1):

                normalizedURL = f"http://members.adult-fanfiction.org/profile.php?no={userID}&view=story&zone={zoneName}&page={pageIndex}"
                soup = self._webSession.GetSoup(normalizedURL)
                if not soup:
                    logging.error(f"Couldn't download page: \"{normalizedURL}\".")
                    return None

                for element in soup.select("div#contentdata > div.alist > ul > li > a"):

                    storyElements.append(element.parent)

        return storyElements

    @staticmethod
    def _GetIDFromURL(URL: str) -> Optional[str]:

        ##
        #
        # Reads the story/user ID from a URL.
        #
        # @param URL The URL.
        #
        # @return The ID; optionally **None**.
        #
        ##

        if not URL:
            return None

        match = re.search(
            "no=(\d+)",
            URL
        )
        if not match:
            return None

        return match.group(1)


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Application.

from fiction_dl.Concepts.Extractor import Extractor
from fiction_dl.Extractors.ExtractorAdultFanfiction import ExtractorAdultFanfiction
from fiction_dl.Extractors.ExtractorAH import ExtractorAH
from fiction_dl.Extractors.ExtractorAO3 import ExtractorAO3
from fiction_dl.Extractors.ExtractorAsstrKristen import ExtractorAsstrKristen
from fiction_dl.Extractors.ExtractorFFNet import ExtractorFFNet
from fiction_dl.Extractors.ExtractorFicWad import ExtractorFicWad
from fiction_dl.Extractors.ExtractorHentaiFoundry import ExtractorHentaiFoundry
from fiction_dl.Extractors.ExtractorHPFF import ExtractorHPFF
from fiction_dl.Extractors.ExtractorLiterotica import ExtractorLiterotica
from fiction_dl.Extractors.ExtractorNajlepszaErotyka import ExtractorNajlepszaErotyka
from fiction_dl.Extractors.ExtractorNifty import ExtractorNifty
from fiction_dl.Extractors.ExtractorQuestionableQuesting import ExtractorQuestionableQuesting
from fiction_dl.Extractors.ExtractorQuotev import ExtractorQuotev
from fiction_dl.Extractors.ExtractorRalst import ExtractorRalst
from fiction_dl.Extractors.ExtractorReddit import ExtractorReddit
from fiction_dl.Extractors.ExtractorSamAndJack import ExtractorSamAndJack
from fiction_dl.Extractors.ExtractorSpaceBattles import ExtractorSpaceBattles
from fiction_dl.Extractors.ExtractorSufficientVelocity import ExtractorSufficientVelocity
from fiction_dl.Extractors.ExtractorTextFile import ExtractorTextFile
from fiction_dl.Extractors.ExtractorWhoFic import ExtractorWhoFic
from fiction_dl.Extractors.ExtractorWuxiaWorld import ExtractorWuxiaWorld

# Standard packages.

from typing import Optional

#
#
#
# Functions.
#
#
#

def CreateExtractor(URL: str) -> Optional[Extractor]:

    ##
    #
    # Creates the appropriate extractor for any given URL.
    #
    # @param URL The URL.
    #
    # @return An initialized extractor object, or **None** if no valid extractor has been found.
    #
    ##

    availableExtractors = [

        ExtractorAdultFanfiction(),
        ExtractorAH(),
        ExtractorAO3(),
        ExtractorAsstrKristen(),
        ExtractorFFNet(),
        ExtractorFicWad(),
        ExtractorHentaiFoundry(),
        ExtractorHPFF(),
        ExtractorLiterotica(),
        ExtractorNajlepszaErotyka(),
        ExtractorNifty(),
        ExtractorQuestionableQuesting(),
        ExtractorQuotev(),
        ExtractorRalst(),
        ExtractorReddit(),
        ExtractorSamAndJack(),
        ExtractorSpaceBattles(),
        ExtractorSufficientVelocity(),
        ExtractorWhoFic(),
        ExtractorWuxiaWorld(),

        ExtractorTextFile(),

    ]

    for extractor in availableExtractors:

        if extractor.Initialize(URL):
            return extractor

    return None


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Application.

from fiction_dl.Concepts.Chapter import Chapter
from fiction_dl.Concepts.Extractor import Extractor
from fiction_dl.Concepts.Story import Story
from fiction_dl.Concepts.StoryPackage import StoryPackage
from fiction_dl.Core.Cache import Cache
from fiction_dl.Core.InputData import InputData
from fiction_dl.Extractors.ExtractorTextFile import ExtractorTextFile
from fiction_dl.Formatters.FormatterEPUB import FormatterEPUB
from fiction_dl.Formatters.FormatterHTML import FormatterHTML
from fiction_dl.Formatters.FormatterMOBI import FormatterMOBI
from fiction_dl.Formatters.FormatterODT import FormatterODT
from fiction_dl.Formatters.FormatterPDF import FormatterPDF
from fiction_dl.Processors.SanitizerProcessor import SanitizerProcessor
from fiction_dl.Processors.TypographyProcessor import TypographyProcessor
from fiction_dl.Utilities.Extractors import CreateExtractor
from fiction_dl.Utilities.General import RenderPDFPageToBytes
from fiction_dl.Utilities.HTML import FindImagesInCode, MakeURLAbsolute
from fiction_dl.Utilities.Text import GetPrintableStoryTitle, Transliterate
import fiction_dl. Configuration as Configuration

# Standard packages.

from argparse import Namespace
import logging
from os.path import expandvars, isfile
from pathlib import Path
import re
from requests.exceptions import ConnectionError
from ssl import SSLError
from time import sleep
from typing import Dict, List, Optional, Union
from urllib3.exceptions import ProtocolError

# Non-standard packages.

from cloudscraper.exceptions import CloudflareChallengeError
from dreamy_utilities.Containers import RemoveDuplicates
from dreamy_utilities.Filesystem import FindExecutable, GetSanitizedFileName, WriteTextFile
from dreamy_utilities.Interface import Interface
from dreamy_utilities.Text import GetCurrentDate, Stringify, Truncate
from dreamy_utilities.Web import GetSiteURL

#
#
#
# Classes.
#
#
#

##
#
# The "Application" class encapsulates the whole logic of the application.
#
##

class Application:

    def __init__(self, arguments: Namespace, cacheDirectoryPath: Path) -> None:

        ##
        #
        # The constructor.
        #
        # @param cacheDirectoryPath The path to the cache directory.
        #
        ##

        self._arguments = arguments

        self._cache = Cache(cacheDirectoryPath)
        self._interface = Interface()

    def Launch(self) -> None:

        ##
        #
        # Launches the application.
        #
        ##

        # Welcome the user and clear the cache.

        self._interface.Text(Configuration.WelcomingMessage, bold = True)

        if self._arguments.ClearCache:
            logging.info("Deleting the cache...")
            self._cache.Clear()

        # Print notices.

        if (notices := self._GenerateNotices()):

            self._interface.EmptyLine()

            for notice in notices:
                self._interface.Notice(notice)

        # Process the input arguments.

        self._interface.Process("Processing input arguments...", section = True)

        inputData = InputData(self._arguments.Input)
        if not self._arguments.Pack:
            inputData.ExpandAndShuffle()
        else:
            inputData.Expand()

        URLs = inputData.Access()

        self._interface.Comment(f"The list contains {len(URLs)} item(s).")

        # Process the stories.

        downloadedStories = []
        skippedURLs = []

        for index, URL in enumerate(URLs, start = 1):

            self._interface.LineBreak()
            self._interface.Text(f'{index}/{len(URLs)}: "{URL}".', section = True, bold = True)

            newlyDownloadedStory = None

            if not self._arguments.Debug:

                try:

                    sleep(Configuration.PostChapterSleepTime)

                    newlyDownloadedStory = self._ProcessURL(URL)

                except KeyboardInterrupt:

                    self._interface.ClearLine()
                    self._interface.Notice("Quitting...")

                    exit()

                except ConnectionError as caughtException:

                    self._interface.Error(f"The website has refused connection: {caughtException}")
                    self._interface.GrabUserAttention()

                except SSLError as caughtException:

                    self._interface.Error(f"An SSL error has occurred: {caughtException}")
                    self._interface.GrabUserAttention()

                except CloudflareChallengeError as caughtException:

                    self._interface.Error("A Cloudflare challenge error has occurred. Try again later.")
                    self._interface.GrabUserAttention()

                except BaseException as caughtException:

                    self._interface.Error(f"An exception has been thrown: {caughtException}")
                    self._interface.GrabUserAttention()

                except:

                    self._interface.Error("An exception has been thrown.")
                    self._interface.GrabUserAttention()

            else:

                sleep(Configuration.PostChapterSleepTime)

                newlyDownloadedStory = self._ProcessURL(URL)

            if not newlyDownloadedStory:
                skippedURLs.append(URL)

            if newlyDownloadedStory:

                if not self._arguments.Pack:

                    if not self._arguments.Debug:

                        try:

                            self._FormatAndSaveStoryOrPackage(newlyDownloadedStory)

                        except FileNotFoundError as caughtException:

                            self._interface.Error(f"A filesystem exception has occurred: {caughtException}")
                            self._interface.GrabUserAttention()
                            skippedURLs.append(URL)

                        except:

                            self._interface.Error("An exception has been thrown.")
                            self._interface.GrabUserAttention()
                            skippedURLs.append(URL)

                    else:

                        self._FormatAndSaveStoryOrPackage(newlyDownloadedStory)

                else:

                    downloadedStories.append(newlyDownloadedStory)

        self._interface.LineBreak()

        # Print information about skipped stories.

        if skippedURLs:

            print()

            for URL in skippedURLs:
                self._interface.Error(f'Failed to download a story: "{URL}".')

            WriteTextFile(Configuration.SkippedURLsFilePath, "\n".join(skippedURLs))

        # Print some final information.

        successCount = len(URLs) - len(skippedURLs)

        self._interface.Comment(f"Downloaded {successCount}/{len(URLs)} stories.", section = True)

        # Save downloaded stories.

        if self._arguments.Pack and downloadedStories:
            self._FormatAndSaveStoryOrPackage(StoryPackage(downloadedStories))

        # Clear the cache.

        if not self._arguments.PersistentCache:
            logging.info("Deleting the cache...")
            self._cache.Clear()

    def _ProcessURL(self, URL: str) -> Optional[Story]:

        ##
        #
        # Processes a URL, in text mode.
        #
        # @param URL The URL to be processed.
        #
        # @return The Story object if the URL has been processed successfully, **None** otherwise.
        #
        ##

        # Locate a working extractor.

        self._interface.Process("Creating the extractor...", section = True)

        extractor = CreateExtractor(URL)
        if not extractor:
            logging.error("No matching extractor found.")
            return None

        self._interface.Comment(f'Extractor created: "{type(extractor).__name__}".')

        # Authenticate the user (if supported by the extractor).

        if self._arguments.Authenticate and extractor.SupportsAuthentication():

            self._interface.Process("Logging-in...", section = True)

            authenticationResult = extractor.Authenticate(self._interface)

            if Extractor.AuthenticationResult.FAILURE == authenticationResult:
                self._interface.Error("Failed to authenticate.")
            elif Extractor.AuthenticationResult.ABANDONED == authenticationResult:
                self._interface.Comment("Proceeding without logging-in...")
            else:
                self._interface.Comment("Authenticated successfully.")

        # Scan the story.

        self._interface.Process("Scanning the story...", section = True)

        if not extractor.ScanStory():
            logging.error("Failed to scan the story.")
            return None

        self._PrintMetadata(extractor.Story)

        # Check whether the output files already exist.

        outputFilePaths = self._GetOutputPaths(self._arguments.Output, extractor.Story)

        if (not self._arguments.Force) and all(x.is_file() for x in outputFilePaths.values()):
            self._interface.Comment("This story has been downloaded already.", section = True)
            return True

        elif self._arguments.Force:
            [x.unlink() for x in outputFilePaths.values() if x.is_file()]

        # Extract content.

        self._interface.Process("Extracting content...", section = True)

        for index in range(1, extractor.Story.Metadata.ChapterCount + 1):

            # Generate cache identifiers.

            cacheOwnerName = extractor.Story.Metadata.URL
            cacheTitleName = f"{index}-Title"
            cacheContentName = f"{index}-Content"

            # Retrieve chapter data, either from cache or by downloading it.

            retrievedFromCache = False

            chapter = Chapter(
                title = Stringify(self._cache.RetrieveItem(cacheOwnerName, cacheTitleName)),
                content = Stringify(self._cache.RetrieveItem(cacheOwnerName, cacheContentName))
            )

            if chapter:

                retrievedFromCache = True

            else:

                chapter = extractor.ExtractChapter(index)

                if not chapter:

                    if (1 != index) and (extractor.Story.Metadata.ChapterCount != index):
                        logging.error("Failed to extract story content.")
                        return None

                    else:
                        self._interface.Error("Failed to extract the last chapter - it doesn't seem to exist.")
                        continue

            extractor.Story.Chapters.append(chapter)

            # Add the chapter to cache.

            if not retrievedFromCache:
                self._cache.AddItem(cacheOwnerName, cacheTitleName, chapter.Title)
                self._cache.AddItem(cacheOwnerName, cacheContentName, chapter.Content)

            # Notify the user, then sleep for a while.

            self._interface.ProgressBar(
                index,
                extractor.Story.Metadata.ChapterCount,
                Configuration.ProgressBarLength,
                f"# Extracted chapter {index}/{extractor.Story.Metadata.ChapterCount}",
                True
            )

            if extractor.Story.Metadata.ChapterCount == index:
                self._interface.EmptyLine()

            if not retrievedFromCache and extractor.RequiresBreaksBetweenRequests():
                sleep(Configuration.PostChapterSleepTime)

        # Locate and download images.

        if self._arguments.Images:

            self._interface.Process("Downloading images...", section = True)

            # Locate the images.

            for chapter in extractor.Story.Chapters:
                extractor.Story.Images.extend(FindImagesInCode(chapter.Content))

            storySiteURL = GetSiteURL(extractor.Story.Metadata.URL)
            for image in extractor.Story.Images:
                image.URL = MakeURLAbsolute(image.URL, storySiteURL)

            self._interface.Comment(f"Found {len(extractor.Story.Images)} image(s).")

            # Download them.

            if extractor.Story.Images:

                imageCount = len(extractor.Story.Images)
                downloadedImageCount = 0

                previousImageFailedToDownload = False

                for index, image in enumerate(extractor.Story.Images, start = 1):

                    retrievedFromCache = False
                    imageData = self._cache.RetrieveItem(extractor.Story.Metadata.URL, image.URL)

                    if not image.CreateFromData(imageData, Configuration.MaximumImageSideLength):

                        imageData = extractor.ExtractMedia(image.URL)

                        if imageData:
                            image.CreateFromData(imageData, Configuration.MaximumImageSideLength)

                    else:

                        retrievedFromCache = True

                    if image:

                        if not retrievedFromCache:
                            self._cache.AddItem(
                                extractor.Story.Metadata.URL,
                                image.URL,
                                image.Data
                            )

                        self._interface.ProgressBar(
                            index,
                            imageCount,
                            Configuration.ProgressBarLength,
                            f"# Downloaded image {index}/{imageCount}",
                            True
                        )

                        if imageCount == index:
                            print()

                        downloadedImageCount += 1
                        previousImageFailedToDownload = False

                    else:

                        if (index > 1) and (not previousImageFailedToDownload):
                            print()

                        errorMessage =                                                       \
                            f'Failed to download image {index}/{imageCount}: "{image.URL}".' \
                            if not imageData else                                            \
                            f'Failed to process/re-encode image {index}/{imageCount}: "{image.URL}".'

                        self._interface.Error(errorMessage)

                        previousImageFailedToDownload = True

                self._interface.Comment(
                    f"Successfully downloaded {downloadedImageCount}/{imageCount} image(s)."
                )

        # Process content.

        self._interface.Process("Processing content...", section = True)

        extractor.Story.Process()

        for index, chapter in enumerate(extractor.Story.Chapters, start = 1):

            # Store original content.

            if self._arguments.Debug:

                fileName = GetSanitizedFileName(f"{index} - Original.html")
                fileSubdirectoryName = GetSanitizedFileName(extractor.Story.Metadata.Title)

                WriteTextFile(
                    Configuration.DebugDirectoryPath / fileSubdirectoryName / fileName,
                    chapter.Content
                )

            # The sanitizer is used twice - once before any other processing, once after every other
            # processor. The first time is required to clean up the story (remove empty tags and tag
            # trees, for example), the second to guarantee that the story is actually sanitized.

            chapter.Content = SanitizerProcessor().Process(chapter.Content)
            chapter.Content = TypographyProcessor().Process(chapter.Content)
            chapter.Content = SanitizerProcessor().Process(chapter.Content)

            # Store processed content.

            if self._arguments.Debug:

                fileName = GetSanitizedFileName(f"{index} - Processed.html")
                fileSubdirectoryName = GetSanitizedFileName(extractor.Story.Metadata.Title)

                WriteTextFile(
                    Configuration.DebugDirectoryPath / fileSubdirectoryName / fileName,
                    chapter.Content
                )

        if not extractor.Story.Metadata.WordCount:
            extractor.Story.Metadata.WordCount = extractor.Story.CalculateWordCount()

        self._interface.Comment("Content processed.")

        # Return.

        return extractor.Story

    def _FormatAndSaveStoryOrPackage(self, story: Union[Story, StoryPackage]) -> bool:

        # Notify the user.

        notificationText =                              \
            "Formatting and saving the story..."        \
            if not isinstance(story, StoryPackage) else \
            "Formatting and saving downloaded stories..."

        self._interface.Process(notificationText, section = True)

        # Generate output file paths and create the output directory.

        filePaths = self._GetOutputPaths(self._arguments.Output, story)
        filePaths["Directory"].mkdir(parents = True, exist_ok = True)

        # Format and save the stories to HTML.

        self._interface.Comment("Saving as HTML... ", end = "")

        formatter = FormatterHTML(self._arguments.Images)
        if not filePaths["HTML"].is_file():

            if not formatter.FormatAndSave(story, filePaths["HTML"]):
                self._interface.Text("Failed!")

            else:
                self._interface.Text("Done!")

        else:

            self._interface.Text("Output file already exists.")

        # Format and save the stories to ODT.

        self._interface.Comment("Saving as ODT... ", end = "")

        formatter = FormatterODT(self._arguments.Images, isinstance(story, StoryPackage))
        if not filePaths["ODT"].is_file():

            if not formatter.FormatAndSave(story, filePaths["ODT"]):
                self._interface.Text("Failed!")

            else:
                self._interface.Text("Done!")

        else:

            self._interface.Text("Output file already exists.")

        # Format and save the stories to PDF.

        self._interface.Comment("Saving as PDF... ", end = "")

        coverImageData = None

        formatter = FormatterPDF(self._arguments.Images)

        if not self._arguments.LibreOffice.is_file():

            self._interface.Text("This output format is unavailable.")

        elif not filePaths["PDF"].is_file():

            if not formatter.ConvertFromODT(filePaths["ODT"], filePaths["PDF"].parent, self._arguments.LibreOffice):
                self._interface.Text("Failed!")

            else:
                self._interface.Text("Done!")

        else:

            self._interface.Text("Output file already exists.")

        if filePaths["PDF"].is_file():
            coverImageData = RenderPDFPageToBytes(filePaths["PDF"], 0)

        # Format and save the stories to EPUB.

        self._interface.Comment("Saving as EPUB... ", end = "")

        formatter = FormatterEPUB(self._arguments.Images, coverImageData)
        if not filePaths["EPUB"].is_file():

            if not formatter.FormatAndSave(story, filePaths["EPUB"]):
                self._interface.Text("Failed!")

            else:
                self._interface.Text("Done!")

        else:

            self._interface.Text("Output file already exists.")

        # Format and save the story to MOBI.

        self._interface.Comment("Saving as MOBI... ", end = "")

        formatter = FormatterMOBI(self._arguments.Images)

        if not FindExecutable("ebook-convert"):

            self._interface.Text("This output format is unavailable.")

        elif not filePaths["MOBI"].is_file():

            if not formatter.ConvertFromEPUB(filePaths["EPUB"], filePaths["MOBI"].parent):
                self._interface.Text("Failed!")

            else:
                self._interface.Text("Done!")

        else:

            self._interface.Text("Output file already exists.")

        # Return.

        return True

    def _PrintMetadata(self, story: Story) -> None:

        ##
        #
        # Prints a story's metadata.
        #
        # @param story The story.
        #
        ##

        prettifiedMetadata = story.Metadata.GetPrettified()

        wordCount =                                     \
            prettifiedMetadata.WordCount                \
            if "0" != prettifiedMetadata.WordCount else \
            "?"

        self._interface.Table([
            ["Title:", GetPrintableStoryTitle(story)],
            ["Author:", prettifiedMetadata.Author],
            ["Date published:", prettifiedMetadata.DatePublished],
            ["Date updated:", prettifiedMetadata.DateUpdated],
            ["Chapter count:", prettifiedMetadata.ChapterCount],
            ["Word count:", wordCount],
        ])

    def _GetOutputPaths(self, outputPath: str, story: Union[Story, StoryPackage]) -> Dict:

        ##
        #
        # Generates file paths for output files.
        #
        # @param outputPath Directory in which all the application's output is placed.
        # @param story      The story, or story package.
        #
        # @return A dictionary: keys correspond to formats, values to file paths.
        #
        ##

        prettifiedMetadata = story.Metadata.GetPrettified()

        sanitizedTitle = GetPrintableStoryTitle(story)
        sanitizedAuthor = GetSanitizedFileName(prettifiedMetadata.Author)

        outputDirectoryPath = Path(expandvars(outputPath))
        outputDirectoryPath =                                        \
            (outputDirectoryPath / sanitizedAuthor / sanitizedTitle) \
            if not isinstance(story, StoryPackage) else              \
            (outputDirectoryPath)

        return {
            "Directory": outputDirectoryPath,
            "HTML" : outputDirectoryPath / (sanitizedTitle + ".html"),
            "ODT"  : outputDirectoryPath / (sanitizedTitle + ".odt" ),
            "PDF"  : outputDirectoryPath / (sanitizedTitle + ".pdf" ),
            "EPUB" : outputDirectoryPath / (sanitizedTitle + ".epub"),
            "MOBI" : outputDirectoryPath / (sanitizedTitle + ".mobi"),
        }

    def _GenerateNotices(self) -> List[str]:

        notices = []

        if not FindExecutable("ebook-convert"):
            notices.append(
                "\"Calibre\" doesn't seem to be installed on this machine. MOBI output files will not "
                "be generated."
            )

        if not self._arguments.LibreOffice.is_file():
            notices.append(
                "\"LibreOffice\" doesn't seem to be installed on this machine. PDF output files will "
                "not be generated."
            )

        return notices


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Application.

from fiction_dl.Concepts.Chapter import Chapter
from fiction_dl.Concepts.Story import Story
import fiction_dl.Configuration as Configuration

# Standard packages.

from enum import Enum
import logging
import requests
from typing import List, Optional

# Non-standard packages.

from bs4 import BeautifulSoup
from dreamy_utilities.Interface import Interface
from dreamy_utilities.Web import GetHostname
from dreamy_utilities.WebSession import WebSession
from fake_useragent import UserAgent

#
#
#
# Classes.
#
#
#

##
#
# An abstract extractor, meant to be inherited from. Represents a concept of a
# tool capable of handling content coming from a specific source.
#
##

class Extractor:

    ##
    #
    # Represents authentication result.
    #
    ##

    class AuthenticationResult(Enum):
        SUCCESS   = 1
        FAILURE   = 2
        ABANDONED = 3

    def __init__(self) -> None:

        ##
        #
        # The constructor.
        #
        ##

        self.Story = None

        self._webSession = WebSession(UserAgent().random)
        self._chapterURLs = []

        self._downloadStorySoupWhenScanning = True
        self._downloadChapterSoupWhenExtracting = True

        self._chapterParserName = "html.parser"

    def GetSupportedHostnames(self) -> List[str]:

        ##
        #
        # Returns a list of hostnames supposed to be supported by the extractor.
        #
        # @return A list of supported hostnames.
        #
        ##

        return []

    def RequiresBreaksBetweenRequests(self) -> bool:

        ##
        #
        # Does the extractor require the application to sleep between subsequent reqests?
        #
        # @return **True** if it does, **False** otherwise.
        #
        ##

        return True

    def SupportsAuthentication(self) -> bool:

        ##
        #
        # Checks whether the extractor supports user authentication.
        #
        # @return **True** if the site *does* support authentication, **False** otherwise.
        #
        ##

        return False

    def Authenticate(self, interface: Interface) -> AuthenticationResult:

        ##
        #
        # Logs the user in, interactively.
        #
        # @param interface The user interface to be used.
        #
        # @return The result of the authentication attempt.
        #
        ##

        return AuthenticationResult.FAILURE

    def Initialize(self, URL: str) -> bool:

        ##
        #
        # Initializes the extractor with given URL.
        #
        # @param URL The URL in question.
        #
        # @return **True** if the extractor can handle the URL, **False** otherwise.
        #
        ##

        if GetHostname(URL) not in self.GetSupportedHostnames():
            return False

        self.Story = Story(self._GetPrettyStoryURL(URL))

        return True

    def ScanChannel(self, URL: str) -> Optional[List[str]]:

        ##
        #
        # Scans the channel: generates the list of story URLs.
        #
        # @return **None** when the scan fails, a list of story URLs when it doesn't fail.
        #
        ##

        return None

    def ScanStory(self) -> bool:

        ##
        #
        # Scans the story: generates the list of chapter URLs and retrieves the
        # metadata.
        #
        # @return **False** when the scan fails, **True** when it doesn't fail.
        #
        ##

        if self.Story is None:
            logging.error("The extractor isn't initialized.")
            return False

        normalizedURL = self._GetNormalizedStoryURL(self.Story.Metadata.URL)

        soup = None

        if self._downloadStorySoupWhenScanning:

            soup = self._webSession.GetSoup(normalizedURL)

            if not soup:
                logging.error(f'Failed to download tag soup: "{normalizedURL}".')
                return False

        return self._InternallyScanStory(normalizedURL, soup)

    def ExtractChapter(self, index: int) -> Optional[Chapter]:

        ##
        #
        # Extracts specific chapter.
        #
        # @param index The index of the chapter to be extracted.
        #
        # @return **True** if the chapter is extracted correctly, **False** otherwise.
        #
        ##

        if index > len(self._chapterURLs):
            logging.error(
                f"Trying to extract chapter {index}. "
                f"Only {len(self._chapterURLs)} chapter(s) located. "
                f"The story supposedly has {self.Story.Metadata.ChapterCount} chapter(s)."
            )
            return None

        chapterURL = self._chapterURLs[index - 1]

        soup = None

        if self._downloadChapterSoupWhenExtracting:

            soup = self._webSession.GetSoup(chapterURL, self._chapterParserName)
            if not soup:
                logging.error(f'Failed to download tag soup: "{chapterURL}".')
                return None

        return self._InternallyExtractChapter(chapterURL, soup)

    def ExtractMedia(self, URL: str) -> Optional[bytes]:

        ##
        #
        # Extracts binary media (an image, for example).
        #
        # @param URL The URL to be extracted.
        #
        # @return The data extracted, as bytes.
        #
        ##

        if not URL:
            return None

        return self._webSession.Get(URL, text = False, stream = True)

    def _InternallyScanStory(
        self,
        URL: str,
        soup: Optional[BeautifulSoup]
    ) -> bool:

        ##
        #
        # Scans the story: generates the list of chapter URLs and retrieves the
        # metadata.
        #
        # @param URL  The URL of the story.
        # @param soup The tag soup.
        #
        # @return **False** when the scan fails, **True** when it doesn't fail.
        #
        ##

        raise NotImplementedError()

    def _InternallyExtractChapter(
        self,
        URL: str,
        soup: Optional[BeautifulSoup]
    ) -> Optional[Chapter]:

        ##
        #
        # Extracts specific chapter.
        #
        # @param URL  The URL of the page containing the chapter.
        # @param soup The tag soup of the page containing the chapter.
        #
        # @return **True** if the chapter is extracted correctly, **False** otherwise.
        #
        ##

        raise NotImplementedError()

    def _GetNormalizedStoryURL(self, URL: str) -> Optional[str]:

        ##
        #
        # Returns a normalized story URL, i.e. one that can be used for anything.
        #
        # @param URL Input URL (given by the user).
        #
        # @return Normalized URL.
        #
        ##

        return URL

    def _GetPrettyStoryURL(self, URL: str) -> Optional[str]:

        ##
        #
        # Returns a "pretty" story URL, i.e. one that can be used for printing in story content.
        #
        # @param URL Input URL (given by the user).
        #
        # @return Pretty URL.
        #
        ##

        return URL


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Standard packages.

from pathlib import Path
from typing import Optional

# Non-standard packages.

from dreamy_utilities.Filesystem import FindExecutable

#
#
#
# Functions.
#
#
#

def GetLibreOfficeExecutablePath() -> Optional[Path]:

    ##
    #
    # Returns the path to the LibreOffice executable (soffice[.exe]).
    #
    # @return The path to the executable.
    #
    ##

    if (path := FindExecutable("soffice", "LibreOffice", "program")):
        return path

    try:

        import winreg

        registry = winreg.ConnectRegistry(None, winreg.HKEY_LOCAL_MACHINE)
        if not registry:
            return None

        for majorVersion in range(5, 7 + 1):

            for minorVersion in range(0, 9 + 1):

                version = f"{majorVersion}.{minorVersion}"

                try:

                    key = winreg.OpenKey(
                        registry,
                        f"SOFTWARE\\The Document Foundation\\LibreOffice\\{version}\\Capabilities"
                    )

                    value, _ = winreg.QueryValueEx(key, "ApplicationIcon")
                    if not value.endswith(".bin,0"):
                        return None

                    value = value[:-6] + ".exe"

                    path = Path(value)
                    if path.is_file():
                        return path
                    else:
                        return None

                except FileNotFoundError:

                    continue

    except ImportError:

        return None

    return None

def GetPackageDirectory() -> Path:

    ##
    #
    # Returns a path to the main directory of the package. Useful for packed files.
    #
    # @return The path to the package directory.
    #
    ##

    return Path(__file__).parent.parent.absolute()


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Application.

from fiction_dl.Concepts.Chapter import Chapter
from fiction_dl.Concepts.Extractor import Extractor

# Standard packages.

from datetime import datetime
import logging
import re
from typing import List, Optional

# Non-standard packages.

from bs4 import BeautifulSoup
from dreamy_utilities.HTML import ReadElementText
from dreamy_utilities.Text import FindFirstMatch, Stringify
from dreamy_utilities.Web import GetHostname

#
#
#
# Classes.
#
#
#

##
#
# An extractor dedicated for Ralst.com.
#
##

class ExtractorRalst(Extractor):

    def __init__(self) -> None:

        ##
        #
        # The constructor.
        #
        ##

        super().__init__()

        self._chapterParserName = "html5lib"

    def GetSupportedHostnames(self) -> List[str]:

        ##
        #
        # Returns a list of hostnames supposed to be supported by the extractor.
        #
        # @return A list of supported hostnames.
        #
        ##

        return [
            "ralst.com",
        ]

    def _InternallyScanStory(
        self,
        URL: str,
        soup: Optional[BeautifulSoup]
    ) -> bool:

        ##
        #
        # Scans the story: generates the list of chapter URLs and retrieves the
        # metadata.
        #
        # @param URL  The URL of the story.
        # @param soup The tag soup.
        #
        # @return **False** when the scan fails, **True** when it doesn't fail.
        #
        ##

        # Retrieve the metadata.

        metadataElement = soup.select("p")[1]

        metadataText = metadataElement.get_text().strip()
        metadataText.replace(r"\n", " ")
        metadataText = metadataText.split("\n")
        if len(metadataText) < 2:
            logging.error("Invalid metadata text format.")
            return False

        title = metadataText[0]
        author = metadataText[1][3:]

        # Set the metadata.

        self.Story.Metadata.Title = title
        self.Story.Metadata.Author = author

        self.Story.Metadata.DatePublished = "?"
        self.Story.Metadata.DateUpdated = "?"

        self.Story.Metadata.ChapterCount = 1
        self.Story.Metadata.WordCount = 0

        self.Story.Metadata.Summary = "No summary."

        # Set chapter URLs.

        self._chapterURLs = [URL]

        # Return.

        return True

    def _InternallyExtractChapter(
        self,
        URL: str,
        soup: Optional[BeautifulSoup]
    ) -> Optional[Chapter]:

        ##
        #
        # Extracts specific chapter.
        #
        # @param URL  The URL of the page containing the chapter.
        # @param soup The tag soup of the page containing the chapter.
        #
        # @return **True** if the chapter is extracted correctly, **False** otherwise.
        #
        ##

        contentElements = soup.select("p")

        contentElements[0].decompose()
        contentElements[1].decompose()

        contentElements[-1].decompose()
        contentElements[-2].decompose()
        contentElements[-3].decompose()

        return Chapter(
            title = None,
            content = Stringify(soup)
        )


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Application.

from fiction_dl.Concepts.Chapter import Chapter
from fiction_dl.Concepts.Extractor import Extractor

# Standard packages.

from datetime import datetime
import logging
import re
from typing import List, Optional

# Non-standard packages.

from bs4 import BeautifulSoup
from dreamy_utilities.HTML import ReadElementText
from dreamy_utilities.Text import FindFirstMatch, Stringify
from dreamy_utilities.Web import GetHostname

#
#
#
# Classes.
#
#
#

##
#
# An extractor dedicated for SamAndJack.net.
#
##

class ExtractorSamAndJack(Extractor):

    def __init__(self) -> None:

        ##
        #
        # The constructor.
        #
        ##

        super().__init__()

        self._chapterTitles = {}

    def GetSupportedHostnames(self) -> List[str]:

        ##
        #
        # Returns a list of hostnames supposed to be supported by the extractor.
        #
        # @return A list of supported hostnames.
        #
        ##

        return [
            "samandjack.net",
        ]


    def ScanChannel(self, URL: str) -> Optional[List[str]]:

        ##
        #
        # Scans the channel: generates the list of story URLs.
        #
        # @return **None** when the scan fails, a list of story URLs when it doesn't fail.
        #
        ##

        if (not URL) or (GetHostname(URL) not in self.GetSupportedHostnames()):
            return None

        if "viewuser.php" not in URL:
            return None

        soup = self._webSession.GetSoup(URL)
        if not soup:
            logging.error(f"Couldn't download page: \"{URL}\".")
            return None

        # Get the number of pages.

        pageCount = 1

        if (pageLinksElement := soup.select_one("#pagelinks")):

            linkElements = pageLinksElement.select("a")
            if len(linkElements):
                pageCount = len(linkElements) - 1

        # Scan stories on every page.

        storyURLs = []

        for pageIndex in range(1, pageCount + 1):

            currentOffset = (pageIndex - 1) * self._STORIES_PER_PAGE
            currentURL = f"{URL}&offset={currentOffset}"

            soup = self._webSession.GetSoup(currentURL)
            if not soup:
                continue

            for element in soup.select("td.main div.listbox div.title"):

                linkElement = element.select_one("a")
                if (not linkElement) or (not linkElement.has_attr("href")):
                    continue

                storyIDMatch = re.search(
                    "sid=(\d+)",
                    linkElement["href"]
                )

                if not storyIDMatch:
                    continue

                storyID = storyIDMatch.group(1)
                storyURL = f"{self._BASE_URL}viewstory.php?sid={storyID}"

                storyURLs.append(storyURL)

        # Return.

        return storyURLs

    def _InternallyScanStory(
        self,
        URL: str,
        soup: Optional[BeautifulSoup]
    ) -> bool:

        ##
        #
        # Scans the story: generates the list of chapter URLs and retrieves the
        # metadata.
        #
        # @param URL  The URL of the story.
        # @param soup The tag soup.
        #
        # @return **False** when the scan fails, **True** when it doesn't fail.
        #
        ##

        # Extract the metadata.

        titleAuthorElements = soup.select("div#pagetitle a")
        if len(titleAuthorElements) < 2:
            logging.error("Title/author elements not found.")
            return False

        titleElement = titleAuthorElements[0]
        authorElement = titleAuthorElements[1]

        # Read additional metadata and retrieve chapter URLs.

        tableOfContentsURL = f"{URL}&index=1"
        soup = self._webSession.GetSoup(tableOfContentsURL)
        if not soup:
            logging.error(f"Failed to download page: \"{tableOfContentsURL}\".")
            return False

        contentElement = soup.select_one("div.listbox > div.content")
        if not contentElement:
            logging.error("Content element not found in the ToC page.")
            return False

        contentText = contentElement.get_text().strip()
        summary = FindFirstMatch(contentText, "Summary: ([^\n]+)\n")
        datePublished = FindFirstMatch(contentText, "Published:\s*([a-zA-Z]+ \d+, \d+)")
        dateUpdated = FindFirstMatch(contentText, "Updated:\s*([a-zA-Z]+ \d+, \d+)")

        for element in soup.select("div#output > p > b > a"):

            if not element.has_attr("href"):
                continue

            chapterURL = self._GetAdultViewURL(self._BASE_URL + element["href"])

            self._chapterURLs.append(chapterURL)
            self._chapterTitles[chapterURL] = element.get_text().strip()

        if not self._chapterURLs:
            logging.error("No chapters found.")
            return False

        # Set the metadata.

        self.Story.Metadata.Title = titleElement.get_text().strip()
        self.Story.Metadata.Author = authorElement.get_text().strip()

        self.Story.Metadata.DatePublished = self._ReformatDate(datePublished) or "?"
        self.Story.Metadata.DateUpdated = self._ReformatDate(dateUpdated) or "?"

        self.Story.Metadata.ChapterCount = len(self._chapterURLs)
        self.Story.Metadata.WordCount = 0

        self.Story.Metadata.Summary = summary

        # Return.

        return True

    def _InternallyExtractChapter(
        self,
        URL: str,
        soup: Optional[BeautifulSoup]
    ) -> Optional[Chapter]:

        ##
        #
        # Extracts specific chapter.
        #
        # @param URL  The URL of the page containing the chapter.
        # @param soup The tag soup of the page containing the chapter.
        #
        # @return **True** if the chapter is extracted correctly, **False** otherwise.
        #
        ##

        # Extract the content.

        contentElement = soup.select_one("div#story")
        if not contentElement:
            logging.error("Couldn't find the content element.")
            return None

        # Return.

        return Chapter(
            title = self._chapterTitles[URL] if (URL in self._chapterTitles) else None,
            content = Stringify(contentElement.encode_contents())
        )

    def _GetNormalizedStoryURL(self, URL: str) -> Optional[str]:

        ##
        #
        # Returns a normalized story URL, i.e. one that can be used for anything.
        #
        # @param URL Input URL (given by the user).
        #
        # @return Normalized URL.
        #
        ##

        return self._GetAdultViewURL(URL)

    def _GetAdultViewURL(self, URL: str) -> Optional[str]:

        ##
        #
        # Returns a URL leading to the page, this time with adult-mode enabled.
        #
        # @param URL Page URL.
        #
        # @return The adult-proofed URL.
        #
        ##

        if not URL:
            return None

        MINIMUM_WARNING = 1
        MAXIMUM_WARNING = 5

        for warningIndex in reversed(range(MINIMUM_WARNING, MAXIMUM_WARNING + 1)):

            currentURL = URL + f"&ageconsent=ok&warning={warningIndex}"
            soup = self._webSession.GetSoup(currentURL)

            if not soup.select_one("div.errortext"):
                return currentURL

        return URL + "&ageconsent=ok"

    @staticmethod
    def _ReformatDate(date: str) -> Optional[str]:

        ##
        #
        # Reformats long date format to standard format.
        #
        # @param date The input date, as read on the webpage.
        #
        # @return ISO-formatted date. Optionally **None**.
        #
        ##

        if not date:
            return None

        try:

            return datetime.strptime(date, "%b %d, %Y").strftime("%Y-%m-%d")

        except ValueError:

            return None

    _BASE_URL = "http://samandjack.net/fanfics/"
    _STORIES_PER_PAGE = 25